"""
Runs libEnsemble with APOSMM with a PETSc/TAO local optimizer.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_tao_nm.py
   python3 test_persistent_aposmm_tao_nm.py --nworkers 3 --comms local
   python3 test_persistent_aposmm_tao_nm.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi tcp
# TESTSUITE_NPROCS: 4
# TESTSUITE_EXTRA: true

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f

import libensemble.gen_funcs

libensemble.gen_funcs.rc.aposmm_optimizers = 'petsc'
from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f

from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float), ('grad', float, n)],
}

gen_out = [
    ('x', float, n),
    ('x_on_cube', float, n),
    ('sim_id', int),
    ('local_min', bool),
    ('local_pt', bool),
]

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f', 'grad'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'localopt_method': 'nm',
        'lb': np.array([-3, -2]),  # This is only for sampling. TAO_NM doesn't honor constraints.
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {'alloc_f': alloc_f}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 1000}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    print('[Manager]:', H[np.where(H['local_min'])]['x'])
    assert np.sum(~H['local_pt']) > 100, "Had to do at least 100 sample points"
    assert np.sum(H['local_pt']) > 100, "Why didn't at least 100 local points occur?"



"""
libensemble helpers for launching subprocesses.
====================================================
"""

import os
import sys
import shlex
import signal
import time
import subprocess

from itertools import chain


def form_command(cmd_template, specs):
    "Fill command parts with dict entries from specs; drop any missing."
    specs = {k: v for k, v in specs.items() if v is not None}

    def fill(fmt):
        "Fill a template string and split with shlex; drop if missing specs"
        try:
            return shlex.split(fmt.format(**specs), posix=False)
        except KeyError:
            return None
    return list(chain.from_iterable(filter(None, map(fill, cmd_template))))


def launch(cmd_template, specs=None, **kwargs):
    "Launch a new subprocess (with command templating and Python 3 help)."
    cmd = (form_command(cmd_template, specs) if specs is not None
           else cmd_template)
    return subprocess.Popen(cmd, **kwargs)


def killpg(process):
    "Kill the process (and group if it is group leader)."
    try:
        pid = process.pid
        pgid = os.getpgid(pid) if hasattr(os, 'killpg') else -1
        if pgid == pid:
            os.killpg(pgid, signal.SIGKILL)
        else:
            process.kill()
        return True
    except OSError:  # In Python 3: ProcessLookupError
        return False


def terminatepg(process):
    "Send termination signal to the process (and group if it is group leader)"
    try:
        pid = process.pid
        pgid = os.getpgid(pid) if hasattr(os, 'killpg') else -1
        if pgid == pid:
            os.killpg(pgid, signal.SIGTERM)
        else:
            process.terminate()
        return True
    except OSError:  # In Python 3: ProcessLookupError
        return False


def process_is_stopped(process, timeout):
    "Wait for timeout to see if process is finished; True if done."
    start_time = time.time()
    while time.time() - start_time < timeout:
        time.sleep(0.01)
        if process.poll() is not None:
            return True
    return process.poll() is not None


# Python 3.3 added timeout arguments
def wait_py32(process, timeout=None):
    "Wait on a process with timeout."
    if timeout is None or process_is_stopped(process, timeout):
        return process.wait()
    return None


def wait_py33(process, timeout=None):
    "Wait on a process with timeout (wait forever if None)."
    try:
        return process.wait(timeout=timeout)
    except subprocess.TimeoutExpired:
        return None


wait = wait_py33 if sys.version_info[0:2] > (3, 2) else wait_py32


def wait_and_kill(process, timeout):
    "Give a grace period for a process to terminate, then kill it."
    rc = wait(process, timeout)
    if rc is not None:
        return rc
    killpg(process)
    return process.wait()


def cancel(process, timeout=0):
    "Send a termination signal, give a grace period, then hard kill if needed."
    if timeout is not None and timeout > 0:
        terminatepg(process)
    return wait_and_kill(process, timeout)


# Note: cancel with timeout 0    -- just kill and then wait
#       cancel with timeout None -- just terminate and then wait
#       cancel with timeout > 0  -- try terminating, then hard kill if needed



"""
This module detects and returns system resources

"""

import os
import socket
import logging
from libensemble.resources import node_resources
from libensemble.resources.mpi_resources import get_MPI_runner
from libensemble.resources.env_resources import EnvResources
from libensemble.resources.worker_resources import (WorkerResources,
                                                    ResourceManager)

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


class ResourcesException(Exception):
    "Resources module exception."


class Resources:
    """Provides system resources to libEnsemble and executor.

    A resources instance is always initialized unless ``libE_specs['disable_resource_manager']`` is ``True``.

    **Class Attributes:**

    :cvar Resources: resources: The resources object is stored here and can be retrieved in user functions.

    **Object Attributes:**

    These are set on initialization.

    :ivar string top_level_dir: Directory where searches for node_list file.
    :ivar GlobalResources glob_resources: Maintains resources available to libEnsemble.

    The following are set up after manager/worker fork.

    The resource manager is set up only on the manaager, while the worker resources object is set
    up on workers.

    :ivar ResourceManager resource_manager: An object that manages resource set assignment to workers.
    :ivar WorkerResources worker_resources: An object that contains worker specific resources.
    """

    resources = None

    DEFAULT_NODEFILE = 'node_list'

    @classmethod
    def init_resources(cls, libE_specs):
        """Initiate resource management"""

        # If disable_resource_manager is True, then Resources.resources will remain None.
        disable_resource_manager = libE_specs.get('disable_resource_manager', False)
        if not disable_resource_manager:
            top_level_dir = os.getcwd()
            if Resources.resources is None:
                Resources.resources = Resources(libE_specs=libE_specs, top_level_dir=top_level_dir)

    def __init__(self, libE_specs, top_level_dir=None):
        """ Initiate a new resources object """
        self.top_level_dir = top_level_dir or os.getcwd()
        self.glob_resources = GlobalResources(libE_specs=libE_specs, top_level_dir=None)
        self.resource_manager = None  # For Manager
        self.worker_resources = None  # For Workers

    def set_worker_resources(self, num_workers, workerid):
        """Initiate the worker resources component of resources"""
        self.worker_resources = WorkerResources(num_workers, self.glob_resources, workerid)

    def set_resource_manager(self, num_workers):
        """Initiate the resource manager component of resources"""
        self.resource_manager = ResourceManager(num_workers, self.glob_resources)

    def add_comm_info(self, libE_nodes):
        """Adds comms-specific information to resources

        Removes libEnsemble nodes from nodelist if in dedicated_mode.
        """
        self.glob_resources.add_comm_info(libE_nodes)


class GlobalResources:
    """
    **Object Attributes:**

    These are set on initialization.
    :ivar string top_level_dir: Directory where searches for node_list file
    :ivar EnvResources env_resources: Object storing environment variables used by resources
    :ivar list global_nodelist: list of all nodes available for running user applications
    :ivar int logical_cores_avail_per_node: Logical cores (including SMT threads) available on a node
    :ivar int physical_cores_avail_per_node: Physical cores available on a node
    :ivar list zero_resource_workers: List of workerIDs to have no resources.
    :ivar boolean dedicated_mode: Whether to remove libE nodes from global nodelist.
    :ivar int num_resource_sets: Number of resource sets, if supplied by the user.
    """

    def __init__(self, libE_specs, top_level_dir=None):

        """Initializes a new Resources instance

        Determines the compute resources available for current allocation, including
        node list and cores/hardware threads available within nodes.

        The following parameters may be extracted from ``libE_specs``

        Parameters
        ----------

        top_level_dir: string, optional
            Directory libEnsemble runs in (default is current working directory)

        dedicated_mode: boolean, optional
            If true, then dedicate nodes to running libEnsemble.
            Dedicated mode means that any nodes running libE processes (manager and workers),
            will not be available to worker launched tasks (user applications). They will
            be removed from the nodelist (if present), before dividing into resource sets.

        zero_resource_workers: list of ints, optional
            List of workers that require no resources.

        num_resource_sets: int, optional
            The total number of resource sets. Resources will be divided into this number.
            Default: None. If None, resources will be divided by workers (excluding zero_resource_workers).

        cores_on_node: tuple (int,int), optional
            If supplied gives (physical cores, logical cores) for the nodes. If not supplied,
            this will be auto-detected.

        enforce_worker_core_bounds: boolean, optional
            If True, then libEnsemble's executor will raise an exception if it detects that
            a worker has been instructed to launch tasks with the number of requested processes
            being excessive to the number of cores allocated to that worker, or not enough
            processes were requested to satisfy allocated cores.

        node_file: String, optional
            If supplied, give the name of a file in the run directory to use as a node-list
            for use by libEnsemble. Defaults to a file named 'node_list'. If the file does
            not exist, then the node-list will be auto-detected.

        nodelist_env_slurm: String, optional
            The environment variable giving a node list in Slurm format (Default: uses SLURM_NODELIST).
            Note: This is queried only if a node_list file is not provided.

        nodelist_env_cobalt: String, optional
            The environment variable giving a node list in Cobalt format (Default: uses COBALT_PARTNAME).
            Note: This is queried only if a node_list file is not provided.

        nodelist_env_lsf: String, optional
            The environment variable giving a node list in LSF format (Default: uses LSB_HOSTS).
            Note: This is queried only if a node_list file is not provided.

        nodelist_env_lsf_shortform: String, optional
            The environment variable giving a node list in LSF short-form format (Default: uses LSB_MCPU_HOSTS)
            Note: This is only queried if a node_list file is not provided.

        """
        self.top_level_dir = top_level_dir
        self.dedicated_mode = libE_specs.get('dedicated_mode', False)
        self.zero_resource_workers = libE_specs.get('zero_resource_workers', [])
        self.num_resource_sets = libE_specs.get('num_resource_sets', None)
        self.enforce_worker_core_bounds = libE_specs.get('enforce_worker_core_bounds', False)

        if self.dedicated_mode:
            logger.debug('Running in central mode')

        resource_info = libE_specs.get('resource_info', {})
        cores_on_node = resource_info.get('cores_on_node', None)
        node_file = resource_info.get('node_file', None)
        nodelist_env_slurm = resource_info.get('nodelist_env_slurm', None)
        nodelist_env_cobalt = resource_info.get('nodelist_env_cobalt', None)
        nodelist_env_lsf = resource_info.get('nodelist_env_lsf', None)
        nodelist_env_lsf_shortform = resource_info.get('nodelist_env_lsf_shortform', None)

        self.env_resources = EnvResources(nodelist_env_slurm=nodelist_env_slurm,
                                          nodelist_env_cobalt=nodelist_env_cobalt,
                                          nodelist_env_lsf=nodelist_env_lsf,
                                          nodelist_env_lsf_shortform=nodelist_env_lsf_shortform)

        if node_file is None:
            node_file = Resources.DEFAULT_NODEFILE

        self.global_nodelist = \
            GlobalResources.get_global_nodelist(node_file=node_file,
                                                rundir=self.top_level_dir,
                                                env_resources=self.env_resources)

        self.shortnames = GlobalResources.is_nodelist_shortnames(self.global_nodelist)
        if self.shortnames:
            self.local_host = self.env_resources.shortnames([socket.gethostname()])[0]
        else:
            self.local_host = socket.gethostname()

        # Note: Launcher used here just to get cores on node etc - independent of whether using MPIExecutor
        self.launcher = get_MPI_runner()
        remote_detect = False
        if self.local_host not in self.global_nodelist:
            remote_detect = True

        if not cores_on_node:
            cores_on_node = \
                node_resources.get_sub_node_resources(launcher=self.launcher,
                                                      remote_mode=remote_detect,
                                                      env_resources=self.env_resources)
        self.physical_cores_avail_per_node = cores_on_node[0]
        self.logical_cores_avail_per_node = cores_on_node[1]
        self.libE_nodes = None

    def add_comm_info(self, libE_nodes):
        """Adds comms-specific information to resources

        Removes libEnsemble nodes from nodelist if in dedicated_mode.
        """
        if self.shortnames:
            self.libE_nodes = self.env_resources.shortnames(libE_nodes)
        else:
            self.libE_nodes = libE_nodes
        libE_nodes_in_list = list(filter(lambda x: x in self.libE_nodes, self.global_nodelist))
        if libE_nodes_in_list:
            if self.dedicated_mode and len(self.global_nodelist) > 1:
                self.global_nodelist = GlobalResources.remove_nodes(self.global_nodelist, self.libE_nodes)
                if not self.global_nodelist:
                    logger.warning("Warning. Node-list for tasks is empty. Remove dedicated_mode or add nodes")

    @staticmethod
    def is_nodelist_shortnames(nodelist):
        """Returns True if any entry contains a '.', else False"""
        for item in nodelist:
            if '.' in item:
                return False
        return True

    # This is for central mode where libE nodes will not share with app nodes
    @staticmethod
    def remove_nodes(global_nodelist_in, remove_list):
        """Removes any nodes in remove_list from the global nodelist"""
        global_nodelist = list(filter(lambda x: x not in remove_list, global_nodelist_in))
        return global_nodelist

    @staticmethod
    def get_global_nodelist(node_file=Resources.DEFAULT_NODEFILE,
                            rundir=None,
                            env_resources=None):
        """
        Returns the list of nodes available to all libEnsemble workers.

        If a node_file exists this is used, otherwise the environment
        is interrogated for a node list. If a dedicated manager node is used,
        then a node_file is recommended.

        In central mode, any node with a libE worker is removed from the list.
        """
        top_level_dir = rundir or os.getcwd()
        node_filepath = os.path.join(top_level_dir, node_file)
        global_nodelist = []
        if os.path.isfile(node_filepath):
            logger.debug("node_file found - getting nodelist from node_file")
            with open(node_filepath, 'r') as f:
                for line in f:
                    global_nodelist.append(line.rstrip())
        else:
            logger.debug("No node_file found - searching for nodelist in environment")
            if env_resources:
                global_nodelist = env_resources.get_nodelist()

            if not global_nodelist:
                # Assume a standalone machine
                logger.info("Can not find nodelist from environment. Assuming standalone")
                # global_nodelist.append(env_resources.shortnames([socket.gethostname()])[0])
                global_nodelist.append(socket.gethostname())

        if global_nodelist:
            return global_nodelist
        raise ResourcesException("Error. global_nodelist is empty")



import numpy as np
import matplotlib.pyplot as plt
from libensemble.libE import libE
from libensemble.tools import add_unique_random_streams
from tutorial_gen import gen_random_sample
from tutorial_sim import sim_find_sine

nworkers = 4
libE_specs = {'nworkers': nworkers, 'comms': 'local'}

gen_specs = {'gen_f': gen_random_sample,        # Our generator function
             'out': [('x', float, (1,))],       # gen_f output (name, type, size).
             'user': {'lower': np.array([-3]),  # random sampling lower bound
                      'upper': np.array([3]),   # random sampling upper bound
                      'gen_batch_size': 5       # number of values gen_f will generate per call
                      }
             }

sim_specs = {'sim_f': sim_find_sine,            # Our simulator function
             'in': ['x'],                       # Input field names. 'x' from gen_f output
             'out': [('y', float)]}             # sim_f output. 'y' = sine('x')

persis_info = add_unique_random_streams({}, nworkers+1)  # Initialize manager/workers random streams

exit_criteria = {'sim_max': 80}                 # Stop libEnsemble after 80 simulations

H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info,
                            libE_specs=libE_specs)

# Some (optional) statements to visualize our History array
print([i for i in H.dtype.fields])
print(H)

colors = ['b', 'g', 'r', 'y', 'm', 'c', 'k', 'w']

for i in range(1, nworkers + 1):
    worker_xy = np.extract(H['sim_worker'] == i, H)
    x = [entry.tolist()[0] for entry in worker_xy['x']]
    y = [entry for entry in worker_xy['y']]
    plt.scatter(x, y, label='Worker {}'.format(i), c=colors[i-1])

plt.title('Sine calculations for a uniformly sampled random distribution')
plt.xlabel('x')
plt.ylabel('sine(x)')
plt.legend(loc='lower right')
plt.savefig('tutorial_sines.png')



from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def give_sim_work_first(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function gives (in order) entries in ``H`` to idle workers
    to evaluate in the simulation function. The fields in ``sim_specs['in']``
    are given. If all entries in `H` have been given a be evaluated, a worker
    is told to call the generator function, provided this wouldn't result in
    more than ``alloc_specs['user']['num_active_gen']`` active generators.

    This fast_alloc variation of give_sim_work_first is useful for cases that
    simply iterate through H, issuing evaluations in order and, in particular,
    is likely to be faster if there will be many short simulation evaluations,
    given that this function contains fewer column length operations.

    .. seealso::
        `test_fast_alloc.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_fast_alloc.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)

    gen_count = support.count_gens()
    Work = {}
    gen_in = gen_specs.get('in', [])

    for wid in support.avail_worker_ids():
        # Skip any cancelled points
        while persis_info['next_to_give'] < len(H) and H[persis_info['next_to_give']]['cancel_requested']:
            persis_info['next_to_give'] += 1

        # Give sim work if possible
        if persis_info['next_to_give'] < len(H):
            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], [persis_info['next_to_give']], [])
            except InsufficientFreeResources:
                break
            persis_info['next_to_give'] += 1

        elif gen_count < user.get('num_active_gens', gen_count + 1):

            # Give gen work
            return_rows = range(len(H)) if gen_in else []
            try:
                Work[wid] = support.gen_work(wid, gen_in, return_rows, persis_info.get(wid))
            except InsufficientFreeResources:
                break
            gen_count += 1
            persis_info['total_gen_calls'] += 1

    return Work, persis_info



import os

outfiles = ['err.txt', 'forces.stat', 'out.txt']


def check_log_exception():
    with open('ensemble.log', 'r') as el:
        out = el.readlines()
    assert 'forces_simf.ForcesException\n' in out, \
        "ForcesException not received by manager or logged."


def test_libe_stats(status):
    with open('libE_stats.txt', 'r') as ls:
        out = ls.readlines()
    assert all([line.endswith(status) for line in out if 'sim' in line]), \
        "Deliberate error status not logged or raised for all sim instances."


def test_ensemble_dir(libE_specs, dir, nworkers, sim_max):
    if not os.path.isdir(dir):
        print('Specified ensemble directory {} not found.'.format(dir))
        return

    if not libE_specs.get('sim_dirs_make'):
        print('Typical tests of ensemble directories dont apply without sim_dirs.')
        return

    if libE_specs.get('use_worker_dirs'):
        assert len(os.listdir(dir)) == nworkers, \
            "Number of worker directories ({}) doesn't match nworkers ({})".format(len(os.listdir(dir)), nworkers)

        num_sim_dirs = 0
        files_found = []

        worker_dirs = [i for i in os.listdir(dir) if i.startswith('worker')]
        for worker_dir in worker_dirs:
            sim_dirs = [i for i in os.listdir(os.path.join(dir, worker_dir)) if i.startswith('sim')]
            num_sim_dirs += len(sim_dirs)

            for sim_dir in sim_dirs:
                files_found.append(all([i in os.listdir(os.path.join(dir, worker_dir, sim_dir))
                                        for i in outfiles]))

        assert num_sim_dirs == sim_max, \
            "Number of simulation specific-directories ({}) doesn't match sim_max ({})".format(num_sim_dirs, sim_max)

        assert all(files_found), \
            "Set of expected files ['err.txt', 'forces.stat', 'out.txt'] not found in each sim_dir."

    else:
        sim_dirs = os.listdir(dir)
        assert all([i.startswith('sim') for i in sim_dirs]), \
            "All directories within ensemble dir not labeled as (or aren't) sim_dirs."

        assert len(sim_dirs) == sim_max, \
            "Number of simulation specific-directories ({}) doesn't match sim_max ({})".format(len(sim_dirs), sim_max)

        files_found = []
        for sim_dir in sim_dirs:
            files_found.append(all([i in os.listdir(os.path.join(dir, sim_dir))
                                    for i in outfiles]))

        assert all(files_found), \
            "Set of expected files ['err.txt', 'forces.stat', 'out.txt'] not found in each sim_dir."

    print('Output directory {} passed tests.'.format(dir))



#!/usr/bin/env python
import os
import numpy as np

from libensemble import Ensemble
from libensemble.executors.mpi_executor import MPIExecutor

####################

sim_app = os.path.join(os.getcwd(), 'forces.x')

if not os.path.isfile('forces.x'):
    if os.path.isfile('build_forces.sh'):
        import subprocess
        subprocess.check_call(['./build_forces.sh'])

####################

forces = Ensemble()
forces.from_yaml('forces.yaml')

forces.logger.set_level('INFO')

if forces.is_manager:
    print('\nRunning with {} workers\n'.format(forces.nworkers))

exctr = MPIExecutor()
exctr.register_app(full_path=sim_app, app_name='forces')

forces.libE_specs['ensemble_dir_path'] = './ensemble'
forces.gen_specs['user'].update({
    'lb': np.array([0]),
    'ub': np.array([32767])
})

forces.persis_info.add_random_streams()

forces.run()

if forces.is_manager:
    forces.save_output(__file__)



"""
This module contains the classes
``Executor`` and ``Task``. The class ``Executor`` can be used to subprocess an application
locally. For MPI programs, or any program using non-local compute resources, one of the
inherited classes should be used. Inherited classes include MPI and Balsam variants.
An ``executor`` can create and manage multiple ``tasks``. The user function
can issue and manage ``tasks`` using the submit, poll, wait, and kill functions.
``Task`` attributes are queried to determine status. Functions are
also provided to access and interrogate files in the ``task``'s working directory. A
``manager_poll`` function can be used to poll for STOP signals from the manager.

"""

import os
import sys
import logging
import itertools
import time

from libensemble.message_numbers import (UNSET_TAG, MAN_SIGNAL_FINISH,
                                         MAN_SIGNAL_KILL, WORKER_DONE,
                                         TASK_FAILED, WORKER_KILL_ON_TIMEOUT,
                                         STOP_TAG)

import libensemble.utils.launcher as launcher
from libensemble.utils.timer import TaskTimer


logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)

STATES = """
UNKNOWN
CREATED
WAITING
RUNNING
FINISHED
USER_KILLED
FAILED""".split()

NOT_STARTED_STATES = '''
CREATED
WAITING
'''.split()

END_STATES = '''
FINISHED
USER_KILLED
FAILED
'''.split()


class ExecutorException(Exception):
    "Raised for any exception in the Executor"


class TimeoutExpired(Exception):
    """Timeout exception raised when Timeout expires"""
    def __init__(self, task, timeout):
        self.task = task
        self.timeout = timeout

    def __str__(self):
        return ("Task {} timed out after {} seconds".format(self.task, self.timeout))


def jassert(test, *args):
    "Version of assert that raises a ExecutorException"
    if not test:
        raise ExecutorException(*args)


class Application:
    """An application is an executable user-program
    (e.g., implementing a sim/gen function)."""

    prefix = 'libe_app'

    def __init__(self, full_path, name=None, calc_type='sim', desc=None):
        """Instantiates a new Application instance."""
        self.full_path = full_path
        self.calc_type = calc_type
        self.calc_dir, self.exe = os.path.split(full_path)

        if self.exe.endswith('.py'):
            self.full_path = ' '.join((sys.executable, full_path))
        self.name = name or self.exe
        self.desc = desc or (self.exe + ' app')
        self.gname = '_'.join([Application.prefix, self.name])


class Task:
    """
    Manages the creation, configuration and status of a launchable task

    """

    prefix = 'libe_task'
    newid = itertools.count()

    def __init__(self, app=None, app_args=None, workdir=None,
                 stdout=None, stderr=None, workerid=None, dry_run=False):
        """Instantiate a new Task instance.

        A new task object is created with an id, status and configuration
        attributes.  This will normally be created by the executor
        on a submission
        """
        self.id = next(Task.newid)

        self.reset()  # Set status attributes
        self.timer = TaskTimer()

        # Run attributes
        self.app = app
        self.app_args = app_args
        self.workerID = workerid

        jassert(app is not None,
                "Task must be created with an app - no app found for task {}".
                format(self.id))

        worker_name = "_worker{}".format(self.workerID) if self.workerID else ""
        self.name = Task.prefix + "_{}{}_{}".format(app.name, worker_name, self.id)
        self.stdout = stdout or self.name + '.out'
        self.stderr = stderr or self.name + '.err'
        self.workdir = workdir
        self.dry_run = dry_run
        self.runline = None
        self.run_attempts = 0

    def reset(self):
        # Status attributes
        self.state = 'CREATED'
        self.process = None
        self.errcode = None
        self.finished = False  # True means task ran, not that it succeeded
        self.success = False
        self.submit_time = None
        self.runtime = 0  # Time since task started to latest poll (or finished).
        self.total_time = None  # Time from task submission until polled as finished.

    def workdir_exists(self):
        """Returns true if the task's workdir exists"""
        return self.workdir and os.path.exists(self.workdir)

    def file_exists_in_workdir(self, filename):
        """Returns true if the named file exists in the task's workdir"""
        return (self.workdir
                and os.path.exists(os.path.join(self.workdir, filename)))

    def read_file_in_workdir(self, filename):
        """Opens and reads the named file in the task's workdir """
        path = os.path.join(self.workdir, filename)
        if not os.path.exists(path):
            raise ValueError("{} not found in working directory".
                             format(filename))
        with open(path) as f:
            return f.read()

    def stdout_exists(self):
        """Returns true if the task's stdout file exists in the workdir"""
        return self.file_exists_in_workdir(self.stdout)

    def read_stdout(self):
        """Opens and reads the task's stdout file in the task's workdir"""
        return self.read_file_in_workdir(self.stdout)

    def stderr_exists(self):
        """Returns true if the task's stderr file exists in the workdir"""
        return self.file_exists_in_workdir(self.stderr)

    def read_stderr(self):
        """Opens and reads the task's stderr file in the task's workdir"""
        return self.read_file_in_workdir(self.stderr)

    def calc_task_timing(self):
        """Calculate timing information for this task"""
        if self.submit_time is None:
            logger.warning("Cannot calc task timing - submit time not set")
            return

        # Do not update if total_time is already set
        if self.total_time is None:
            self.timer.stop()
            self.runtime = self.timer.elapsed
            self.total_time = self.runtime  # For direct launched tasks

    def _check_poll(self):
        """Check whether polling this task makes sense."""
        jassert(self.process is not None,
                "Polled task {} has no process ID - check tasks been launched".
                format(self.name))
        if self.finished:
            logger.debug("Polled task {} has already finished. "
                         "Not re-polling. Status is {}".
                         format(self.name, self.state))
            return False
        return True

    def _set_complete(self, dry_run=False):
        """Set task as complete"""
        self.finished = True
        if dry_run:
            self.success = True
            self.state = 'FINISHED'
        else:
            self.calc_task_timing()
            self.errcode = self.process.returncode
            self.success = (self.errcode == 0)
            self.state = 'FINISHED' if self.success else 'FAILED'
            logger.info("Task {} finished with errcode {} ({})".
                        format(self.name, self.errcode, self.state))

    def poll(self):
        """Polls and updates the status attributes of the task"""
        if self.dry_run:
            return

        if not self._check_poll():
            return

        # Poll the task
        poll = self.process.poll()
        if poll is None:
            self.state = 'RUNNING'
            self.runtime = self.timer.elapsed
            return

        self._set_complete()

    def wait(self, timeout=None):
        """Waits on completion of the task or raises TimeoutExpired exception

        Status attributes of task are updated on completion.

        Parameters
        ----------

        timeout:
            Time in seconds after which a TimeoutExpired exception is raised"""

        if self.dry_run:
            return

        if not self._check_poll():
            return

        # Wait on the task
        rc = launcher.wait(self.process, timeout)
        if rc is None:
            raise TimeoutExpired(self.name, timeout)

        self._set_complete()

    def kill(self, wait_time=60):
        """Kills or cancels the supplied task

        Sends SIGTERM, waits for a period of <wait_time> for graceful
        termination, then sends a hard kill with SIGKILL.  If <wait_time>
        is 0, we go immediately to SIGKILL; if <wait_time> is none, we
        never do a SIGKILL.
        """
        if self.dry_run:
            return

        if self.finished:
            logger.warning("Trying to kill task that is no longer running. "
                           "Task {}: Status is {}".format(self.name, self.state))
            return

        if self.process is None:
            time.sleep(0.2)
            jassert(self.process is not None,
                    "Attempting to kill task {} that has no process ID - "
                    "check tasks been launched".format(self.name))

        logger.info("Killing task {}".format(self.name))
        launcher.cancel(self.process, wait_time)
        self.state = 'USER_KILLED'
        self.finished = True
        self.calc_task_timing()


class Executor:
    """The executor can create, poll and kill runnable tasks

    **Class Attributes:**

    :cvar Executor: executor: The executor object is stored here and can be retrieved in user functions.

    **Object Attributes:**

    :ivar list list_of_tasks: A list of tasks created in this executor

    """

    executor = None

    def _wait_on_start(self, task, fail_time=None):
        '''Called by submit when wait_on_start is True.

        Blocks until task polls as having started.
        If fail_time is supplied, will also block until either task is in an
        end state or fail_time has expired.
        '''
        start = time.time()
        task.timer.start()  # To ensure a start time before poll - will be overwritten unless finished by poll.
        task.submit_time = task.timer.tstart
        while task.state in NOT_STARTED_STATES:
            time.sleep(0.02)
            task.poll()
        logger.debug("Task {} polled as {} after {} seconds".format(task.name, task.state, time.time()-start))
        if not task.finished:
            task.timer.start()
            task.submit_time = task.timer.tstart
            if fail_time:
                remaining = fail_time - task.timer.elapsed
                while task.state not in END_STATES and remaining > 0:
                    time.sleep(min(0.2, remaining))
                    task.poll()
                    remaining = fail_time - task.timer.elapsed
                logger.debug("After {} seconds: task {} polled as {}".format(task.timer.elapsed, task.name, task.state))

    def __init__(self):
        """Instantiate a new Executor instance.

        A new Executor object is created.
        This is typically created in the user calling script.

        """
        self.manager_signal = 'none'
        self.default_apps = {'sim': None, 'gen': None}
        self.apps = {}

        self.wait_time = 60
        self.list_of_tasks = []
        self.workerID = None
        self.comm = None
        Executor.executor = self

    def serial_setup(self):
        """Set up to be called by only one process"""
        pass  # To be overloaded

    @property
    def sim_default_app(self):
        """Returns the default simulation app"""
        return self.default_apps['sim']

    @property
    def gen_default_app(self):
        """Returns the default generator app"""
        return self.default_apps['gen']

    def get_app(self, app_name):
        """Gets the app for a given app_name or raise exception"""
        try:
            app = self.apps[app_name]
        except KeyError:
            app_keys = list(self.apps.keys())
            raise ExecutorException("Application {} not found in registry".format(app_name),
                                    "Registered applications: {}".format(app_keys))
        return app

    def default_app(self, calc_type):
        """Gets the default app for a given calc type"""
        app = self.default_apps.get(calc_type)
        jassert(calc_type in ['sim', 'gen'],
                "Unrecognized calculation type", calc_type)
        jassert(app, "Default {} app is not set".format(calc_type))
        return app

    def set_resources(self, resources):
        # Does not use resources
        pass

    def register_app(self, full_path, app_name=None, calc_type=None, desc=None):
        """Registers a user application to libEnsemble.

        The ``full_path`` of the application must be supplied. Either
        ``app_name`` or ``calc_type`` can be used to identify the
        application in user scripts (in the **submit** function).
        ``app_name`` is recommended.

        Parameters
        ----------

        full_path: String
            The full path of the user application to be registered

        app_name: String, optional
            Name to identify this application.

        calc_type: String, optional
            Calculation type: Set this application as the default 'sim'
            or 'gen' function.

        desc: String, optional
            Description of this application

        """
        if not app_name:
            app_name = os.path.split(full_path)[1]
        self.apps[app_name] = Application(full_path, app_name, calc_type, desc)

        # Default sim/gen apps will be deprecated. Just use names.
        if calc_type is not None:
            jassert(calc_type in self.default_apps,
                    "Unrecognized calculation type", calc_type)
            self.default_apps[calc_type] = self.apps[app_name]

    def manager_poll(self):
        """
        .. _manager_poll_label:

        Polls for a manager signal

        The executor manager_signal attribute will be updated.

        """

        self.manager_signal = 'none'  # Reset

        # Check for messages; disregard anything but a stop signal
        if not self.comm.mail_flag():
            return
        mtag, man_signal = self.comm.recv()
        if mtag != STOP_TAG:
            return

        # Process the signal and push back on comm (for now)
        logger.info('Worker received kill signal {} from manager'.format(man_signal))
        if man_signal == MAN_SIGNAL_FINISH:
            self.manager_signal = 'finish'
        elif man_signal == MAN_SIGNAL_KILL:
            self.manager_signal = 'kill'
        else:
            logger.warning("Received unrecognized manager signal {} - "
                           "ignoring".format(man_signal))
        self.comm.push_to_buffer(mtag, man_signal)
        return man_signal

    def polling_loop(self, task, timeout=None, delay=0.1, poll_manager=False):
        """ Optional, blocking, generic task status polling loop. Operates until the task
        finishes, times out, or is optionally killed via a manager signal. On completion, returns a
        presumptive :ref:`calc_status<datastruct-calc-status>` integer. Potentially useful
        for running an application via the Executor until it stops without monitoring
        its intermediate output.

        Parameters
        ----------

        task: object
            a Task object returned by the executor on submission

        timeout: int, optional
            Maximum number of seconds for the polling loop to run. Tasks that run
            longer than this limit are killed. Default: No timeout

        delay: int, optional
            Sleep duration between polling loop iterations. Default: 0.1 seconds

        poll_manager: bool, optional
            Whether to also poll the manager for 'finish' or 'kill' signals.
            If detected, the task is killed. Default: False.

        Returns
        -------
        calc_status: int
            presumptive integer attribute describing the final status of a launched task

        """

        calc_status = UNSET_TAG

        while not task.finished:
            task.poll()

            if poll_manager:
                man_signal = self.manager_poll()
                if self.manager_signal != 'none':
                    task.kill()
                    calc_status = man_signal
                    break

            if timeout is not None and task.runtime > timeout:
                task.kill()
                calc_status = WORKER_KILL_ON_TIMEOUT
                break

            time.sleep(delay)

        if calc_status == UNSET_TAG:
            if task.state == 'FINISHED':
                calc_status = WORKER_DONE
            elif task.state == 'FAILED':
                calc_status = TASK_FAILED
            else:
                logger.warning("Warning: Task {} in unknown state {}. Error code {}"
                               .format(self.name, self.state, self.errcode))

        return calc_status

    def get_task(self, taskid):
        """ Returns the task object for the supplied task ID """
        task = next((j for j in self.list_of_tasks if j.id == taskid), None)
        if task is None:
            logger.warning("Task {} not found in tasklist".format(taskid))
        return task

    def set_workerID(self, workerid):
        """Sets the worker ID for this executor"""
        self.workerID = workerid

    def set_worker_info(self, comm, workerid=None):
        """Sets info for this executor"""
        self.workerID = workerid
        self.comm = comm

    def submit(self, calc_type=None, app_name=None, app_args=None,
               stdout=None, stderr=None, dry_run=False, wait_on_start=False):
        """Create a new task and run as a local serial subprocess.

        The created task object is returned.

        Parameters
        ----------

        calc_type: String, optional
            The calculation type: 'sim' or 'gen'
            Only used if app_name is not supplied. Uses default sim or gen application.

        app_name: String, optional
            The application name. Must be supplied if calc_type is not.

        app_args: string, optional
            A string of the application arguments to be added to task
            submit command line

        stdout: string, optional
            A standard output filename

        stderr: string, optional
            A standard error filename

        dry_run: boolean, optional
            Whether this is a dry_run - no task will be launched; instead
            runline is printed to logger (at INFO level)

        wait_on_start: boolean, optional
            Whether to wait for task to be polled as RUNNING (or other
            active/end state) before continuing

        Returns
        -------

        task: obj: Task
            The launched task object

        """

        if app_name is not None:
            app = self.get_app(app_name)
        elif calc_type is not None:
            app = self.default_app(calc_type)
        else:
            raise ExecutorException("Either app_name or calc_type must be set")

        default_workdir = os.getcwd()
        task = Task(app, app_args, default_workdir, stdout, stderr, self.workerID)
        runline = task.app.full_path.split()
        if task.app_args is not None:
            runline.extend(task.app_args.split())

        if dry_run:
            logger.info('Test (No submit) Runline: {}'.format(' '.join(runline)))
        else:
            # Launch Task
            logger.info("Launching task {}: {}".
                        format(task.name, " ".join(runline)))
            with open(task.stdout, 'w') as out, open(task.stderr, 'w') as err:
                task.process = launcher.launch(runline, cwd='./',
                                               stdout=out,
                                               stderr=err,
                                               start_new_session=False)
            if (wait_on_start):
                self._wait_on_start(task, 0)  # No fail time as no re-starts in-place

            if not task.timer.timing:
                task.timer.start()
                task.submit_time = task.timer.tstart  # Time not date - may not need if using timer.

            self.list_of_tasks.append(task)
        return task

    def poll(self, task):
        "Polls a task"
        task.poll()

    def kill(self, task):
        "Kills a task"
        jassert(isinstance(task, Task), "Invalid task has been provided")
        task.kill(self.wait_time)



"""
Runs libEnsemble with Latin hypercube sampling on a simple 1D problem

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_1d_sampling.py
   python3 test_1d_sampling.py --nworkers 3 --comms local
   python3 test_1d_sampling.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.one_d_func import one_d_example as sim_f
from libensemble.gen_funcs.sampling import latin_hypercube_sample as gen_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()
libE_specs['save_every_k_gens'] = 300
libE_specs['safe_mode'] = False

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'out': [('x', float, (1,))],
    'user': {
        'gen_batch_size': 500,
        'lb': np.array([-3]),
        'ub': np.array([3]),
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1, seed=1234)

exit_criteria = {'gen_max': 501}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    assert len(H) >= 501
    print("\nlibEnsemble with random sampling has generated enough points")
    save_libE_output(H, persis_info, __file__, nworkers)



"""
Tests various capabilities of the libEnsemble fast_alloc alloc_f

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_fast_alloc.py
   python3 test_fast_alloc.py --nworkers 3 --comms local

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 2 4

import sys
import gc
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel_simple as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.alloc_funcs.fast_alloc import give_sim_work_first as alloc_f
from libensemble.alloc_funcs.only_one_gen_alloc import ensure_one_active_gen as alloc_f2
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

num_pts = 30 * (nworkers)

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float), ('large', float, 1000000)],
    'user': {},
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [('x', float, (2,))],
    'user': {
        'gen_batch_size': num_pts,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 2 * num_pts, 'elapsed_wallclock_time': 300}

if libE_specs['comms'] == 'tcp':
    # Can't use the same interface for manager and worker if we want
    # repeated calls to libE -- the manager sets up a different server
    # each time, and the worker will not know what port to connect to.
    sys.exit("Cannot run with tcp when repeated calls to libE -- aborting...")

for time in np.append([0], np.logspace(-5, -1, 5)):
    print("Starting for time: ", time, flush=True)
    if time == 0:
        alloc_specs = {'alloc_f': alloc_f2, 'out': []}
    else:
        alloc_specs = {'alloc_f': alloc_f, 'out': [], 'user': {'num_active_gens': 1}}

    for rep in range(1):
        sim_specs['user']['pause_time'] = time

        if time == 0:
            sim_specs['user'].pop('pause_time')
            gen_specs['user']['gen_batch_size'] = num_pts // 2

        persis_info['next_to_give'] = 0
        persis_info['total_gen_calls'] = 1

        H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

        if is_manager:
            assert flag == 0
            assert len(H) == 2 * num_pts

        del H
        gc.collect()  # If doing multiple libE calls, users might need to clean up their memory space.



import numpy as np

from tutorial_six_hump_camel import six_hump_camel

from libensemble.libE import libE
from libensemble.gen_funcs.persistent_aposmm import aposmm
from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc
from libensemble.tools import parse_args, add_unique_random_streams

import libensemble.gen_funcs
libensemble.gen_funcs.rc.aposmm_optimizers = 'scipy'

nworkers, is_manager, libE_specs, _ = parse_args()

sim_specs = {'sim_f': six_hump_camel,  # Simulation function
             'in': ['x'],              # Accepts 'x' values
             'out': [('f', float)]}    # Returns f(x) values

gen_out = [('x', float, 2),            # Produces 'x' values
           ('x_on_cube', float, 2),    # 'x' values scaled to unit cube
           ('sim_id', int),            # Produces IDs for sim order
           ('local_min', bool),        # Is a point a local minimum?
           ('local_pt', bool)]         # Is a point from a local opt run?

gen_specs = {'gen_f': aposmm,          # APOSMM generator function
             'persis_in': ['x', 'f', 'x_on_cube', 'sim_id', 'local_min', 'local_pt'],
             'out': gen_out,           # Output defined like above dict
             'user': {'initial_sample_size': 100,  # Random sample 100 points to start
                      'localopt_method': 'scipy_Nelder-Mead',
                      'opt_return_codes': [0],   # Return code specific to localopt_method
                      'max_active_runs': 6,      # Occur in parallel
                      'lb': np.array([-2, -1]),  # Lower bound of search domain
                      'ub': np.array([2, 1])}    # Upper bound of search domain
             }

alloc_specs = {'alloc_f': persistent_aposmm_alloc}

exit_criteria = {'sim_max': 2000}
persis_info = add_unique_random_streams({}, nworkers + 1)

H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info,
                            alloc_specs, libE_specs)
if is_manager:
    print('Minima:', H[np.where(H['local_min'])]['x'])



#!/usr/bin/env python

# ********************* NOT YET IMPLEMENTED ***********************
# ********************* Interface demo: see exctr.manager_poll(task)

# Test of executor module for libensemble:
#  Detecting manager kill signal
#  This keeps MPI out of user code and is portable across different
#  worker concurrency schemes (MPI/threading/multiprocessing)

# Test does not require running full libensemble

import os
from libensemble.executors.executor import Executor


def build_simfunc():
    import subprocess

    # Build simfunc
    # buildstring='mpif90 -o my_simtask.x my_simtask.f90' # On cray need to use ftn
    buildstring = 'mpicc -o my_simtask.x simdir/my_simtask.c'
    # subprocess.run(buildstring.split(),check=True) # Python3.5+
    subprocess.check_call(buildstring.split())


# --------------- Calling script ------------------------------------------

# sim_app = 'simdir/my_simtask.x'
# gen_app = 'gendir/my_gentask.x'

# temp
sim_app = './my_simtask.x'

if not os.path.isfile(sim_app):
    build_simfunc()

USE_BALSAM = False  # Take as arg
# USE_BALSAM = True # Take as arg

# Create and add exes to registry
if USE_BALSAM:
    from libensemble.executors.balsam_executor import BalsamMPIExecutor
    exctr = BalsamMPIExecutor()
else:
    from libensemble.executors.mpi_executor import MPIExecutor
    exctr = MPIExecutor()

exctr.register_app(full_path=sim_app, calc_type='sim')

# Alternative to IF could be using eg. fstring to specify: e.g:
# EXECUTOR = 'Balsam'
# registry = f"{EXECUTOR}Register()"


# --------------- Worker: sim func ----------------------------------------
# Should work with Balsam or not

def polling_loop(exctr, task, timeout_sec=20.0, delay=2.0):
    import time
    start = time.time()

    while time.time() - start < timeout_sec:

        exctr.manager_poll(task)

        if task.manager_signal == 'kill':
            print('Manager has sent kill signal - killing task')
            exctr.kill(task)

        # In future might support other manager signals eg:
        elif task.manager_signal == 'pause':
            # checkpoint_task()
            pass

        time.sleep(delay)
        print('Polling at time', time.time() - start)
        task.poll()
        if task.finished:
            break
        elif task.state == 'WAITING':
            print('Task waiting to execute')
        elif task.state == 'RUNNING':
            print('Task still running ....')

    if task.finished:
        if task.state == 'FINISHED':
            print('Task finished successfully. Status:', task.state)
        elif task.state == 'FAILED':
            print('Task failed. Status:', task.state)
        elif task.state == 'USER_KILLED':
            print('Task has been killed. Status:', task.state)
        else:
            print('Task status:', task.state)
    else:
        print("Task timed out")
        exctr.kill(task)
        if task.finished:
            print('Now killed')
            # double check
            task.poll()
            print('Task state is', task.state)


# Tests
# ********************* NOT YET IMPLEMENTED ***********************

# From worker call Executor by different name to ensure getting registered app from Executor
exctr = Executor.executor

print('\nTest 1 - should complete successfully with status FINISHED :\n')
cores = 4
args_for_sim = 'sleep 5'

task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
polling_loop(exctr, task)

print('\nTest 2 - Task should be MANAGER_KILLED \n')
cores = 4
args_for_sim = 'sleep 5'

task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
polling_loop(exctr, task)



import numpy as np


def EvaluateFunction(theta, component, X, y, c, reg):
    """
    Evaluates linear regression with l2 regularization
    """
    i = component
    m = len(y)

    y_i = y[i]
    X_i = X[:, i]

    base = np.exp(-y_i * np.dot(X_i, theta))

    f_i = (1/m) * np.log(1+base)

    assert reg == 'l2', "Only l2 regularization allowed"
    # if reg is None:
    #     reg_val = 0
    # elif reg == 'l1':
    #     reg_val = (c/m) * np.sum(np.abs(theta))
    reg_val = (c/m) * np.dot(theta, theta)

    return f_i+reg_val


def EvaluateJacobian(theta, component, X, y, c, reg):
    """
    Evaluates linear regression with l2 regularization
    """

    i = component
    m = len(y)

    y_i = y[i]
    X_i = X[:, i]

    base = np.exp(-y_i * np.dot(X_i, theta))

    df_i = (1/m) * (-y_i*base)/(1+base) * X_i

    assert reg == 'l2', "Only l2 regularization allowed"
    # if reg is None:
    #     reg_val = 0
    # elif reg == 'l1':
    #     reg_val = (c/m) * np.sign(theta)
    reg_val = (2*c/m) * theta

    return df_i+reg_val


def logistic_regression_eval(H, persis_info, sim_specs, _):

    X = persis_info['params']['X']
    y = persis_info['params']['y']
    c = persis_info['params']['c']
    reg = persis_info['params'].get('reg', None)

    assert (reg is None) or (reg == 'l1') or (reg == 'l2'), \
        'Incompatable regularization {}'.format(reg)

    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        obj_component = H['obj_component'][i]  # which f_i

        if H[i]['get_grad']:
            H_o['gradf_i'][i] = EvaluateJacobian(x, obj_component, X, y, c, reg)
        else:
            H_o['f_i'][i] = EvaluateFunction(x, obj_component, X, y, c, reg)

    return H_o, persis_info



# !/usr/bin/env python
# Integration Test of executor module for libensemble
# Test does not require running full libensemble
import os
import re
import sys
import time
import pytest
import socket
from libensemble.resources.mpi_resources import MPIResourcesException
from libensemble.executors.executor import Executor, ExecutorException, TimeoutExpired
from libensemble.executors.executor import NOT_STARTED_STATES


USE_BALSAM = False
NCORES = 1
build_sims = ['my_simtask.c', 'my_serialtask.c', 'c_startup.c']

sim_app = 'simdir/my_simtask.x'
serial_app = 'simdir/my_serialtask.x'
c_startup = 'simdir/c_startup.x'
py_startup = 'simdir/py_startup.py'


def setup_module(module):
    try:
        print("setup_module module:%s" % module.__name__)
    except AttributeError:
        print("setup_module (direct run) module:%s" % module)
    if Executor.executor is not None:
        del Executor.executor
        Executor.executor = None
    build_simfuncs()


def setup_function(function):
    print("setup_function function:%s" % function.__name__)
    if Executor.executor is not None:
        del Executor.executor
        Executor.executor = None


def teardown_module(module):
    try:
        print("teardown_module module:%s" % module.__name__)
    except AttributeError:
        print("teardown_module (direct run) module:%s" % module)
    if Executor.executor is not None:
        del Executor.executor
        Executor.executor = None


def build_simfuncs():
    import subprocess
    for sim in build_sims:
        app_name = '.'.join([sim.split('.')[0], 'x'])
        if not os.path.isfile(app_name):
            buildstring = 'mpicc -o ' + os.path.join('simdir', app_name) + ' ' + os.path.join('simdir', sim)
            subprocess.check_call(buildstring.split())


# This would typically be in the user calling script
def setup_executor():
    """Set up an MPI Executor with sim app"""
    if USE_BALSAM:
        from libensemble.executors.balsam_executor import BalsamMPIExecutor
        exctr = BalsamMPIExecutor()
    else:
        from libensemble.executors.mpi_executor import MPIExecutor
        exctr = MPIExecutor()
    exctr.register_app(full_path=sim_app, calc_type='sim')


def setup_serial_executor():
    """Set up serial Executor"""
    from libensemble.executors.executor import Executor
    exctr = Executor()
    exctr.register_app(full_path=serial_app, calc_type='sim')


def setup_executor_startups():
    """Set up serial Executor"""
    from libensemble.executors.executor import Executor
    exctr = Executor()
    exctr.register_app(full_path=c_startup, app_name='c_startup')
    exctr.register_app(full_path=py_startup, app_name='py_startup')


def setup_executor_noapp():
    """Set up an MPI Executor but do not register application"""
    if USE_BALSAM:
        from libensemble.executors.balsam_executor import BalsamMPIExecutor
        exctr = BalsamMPIExecutor()
    else:
        from libensemble.executors.mpi_executor import MPIExecutor
        exctr = MPIExecutor()
        if exctr.workerID is not None:
            sys.exit("Something went wrong in creating Executor")


def setup_executor_fakerunner():
    """Set up an MPI Executor with a non-existent MPI runner"""
    if USE_BALSAM:
        print('Balsom does not support this feature - running MPIExecutor')

    # Create non-existent MPI runner.
    customizer = {'mpi_runner': 'custom',
                  'runner_name': 'non-existent-runner',
                  'subgroup_launch': True}

    from libensemble.executors.mpi_executor import MPIExecutor
    exctr = MPIExecutor(custom_info=customizer)
    exctr.register_app(full_path=sim_app, calc_type='sim')


# -----------------------------------------------------------------------------
# The following would typically be in the user sim_func
def polling_loop(exctr, task, timeout_sec=1, delay=0.05):
    """Iterate over a loop, polling for an exit condition"""
    start = time.time()

    while time.time() - start < timeout_sec:
        time.sleep(delay)

        # Check output file for error
        if task.stdout_exists():
            if 'Error' in task.read_stdout():
                print("Found(deliberate) Error in output file - cancelling task")
                exctr.kill(task)
                time.sleep(delay)  # Give time for kill
                break

        print('Polling at time', time.time() - start)
        task.poll()
        if task.finished:
            break
        elif task.state == 'WAITING':
            print('Task waiting to execute')
        elif task.state == 'RUNNING':
            print('Task still running ....')

    if not task.finished:
        assert task.state == 'RUNNING', "task.state expected to be RUNNING. Returned: " + str(task.state)
        print("Task timed out - killing")
        exctr.kill(task)
        time.sleep(delay)  # Give time for kill
    return task


def polling_loop_multitask(exctr, task_list, timeout_sec=4.0, delay=0.05):
    """Iterate over a loop, polling for exit conditions on multiple tasks"""
    start = time.time()

    while time.time() - start < timeout_sec:

        # Test all done - (return list of not-finished tasks and test if empty)
        active_list = [task for task in task_list if not task.finished]
        if not active_list:
            break

        for task in task_list:
            if not task.finished:
                time.sleep(delay)
                print('Polling task %d at time %f' % (task.id, time.time() - start))
                task.poll()
                if task.finished:
                    continue
                elif task.state == 'WAITING':
                    print('Task %d waiting to execute' % (task.id))
                elif task.state == 'RUNNING':
                    print('Task %d still running ....' % (task.id))

                # Check output file for error
                if task.stdout_exists():
                    if 'Error' in task.read_stdout():
                        print("Found (deliberate) Error in output file - cancelling task %d" % (task.id))
                        exctr.kill(task)
                        time.sleep(delay)  # Give time for kill
                        continue
    return task_list


# Tests ========================================================================================
def test_launch_and_poll():
    """ Test of launching and polling task and exiting on task finish"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 0.2'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    task = polling_loop(exctr, task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)
    assert task.run_attempts == 1, "task.run_attempts should be 1. Returned " + str(task.run_attempts)


def test_launch_and_wait():
    """ Test of launching and waiting on task"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 1'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    task.wait()
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)
    task.wait()  # Already complete
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)


def test_launch_and_wait_timeout():
    """ Test of launching and waiting on task timeout (and kill)"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 5'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    try:
        task.wait(timeout=0.5)
    except TimeoutExpired:
        assert not task.finished, "task.finished should be False. Returned " + str(task.finished)
        task.kill()
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'USER_KILLED', "task.state should be USER_KILLED. Returned " + str(task.state)


def test_launch_wait_on_start():
    """ Test of launching task with wait_on_start """
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 0.2'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim, wait_on_start=True)
    assert task.state not in NOT_STARTED_STATES, "Task should not be in a NOT_STARTED state. State: " + str(task.state)
    exctr.poll(task)
    if not task.finished:
        task = polling_loop(exctr, task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)


def test_kill_on_file():
    """ Test of killing task based on something in output file"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 0.1 Error'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    task = polling_loop(exctr, task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'USER_KILLED', "task.state should be USER_KILLED. Returned " + str(task.state)


def test_kill_on_timeout():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 10'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    task = polling_loop(exctr, task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'USER_KILLED', "task.state should be USER_KILLED. Returned " + str(task.state)


def test_kill_on_timeout_polling_loop_method():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 10'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    exctr.polling_loop(task, timeout=1)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'USER_KILLED', "task.state should be USER_KILLED. Returned " + str(task.state)


def test_launch_and_poll_multitasks():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    task_list = []
    cores = NCORES

    for j in range(3):
        # outfilename = 'out_' + str(j) + '.txt' # Could allow launch to generate outfile names based on task.id
        outfile = 'multitask_task_' + str(j) + '.out'
        sleeptime = 0.3 + (j*0.2)  # Change args
        args_for_sim = 'sleep' + ' ' + str(sleeptime)
        # rundir = 'run_' + str(sleeptime)
        task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim, stdout=outfile)
        task_list.append(task)

    task_list_return = polling_loop_multitask(exctr, task_list)
    for task in task_list_return:
        assert task.finished, "task.finished should be True. Returned " + str(task.finished)
        assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)


def test_get_task():
    """Return task from given task id"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor

    # Try with no tasks set up
    A = exctr.get_task('a')
    assert A is None, 'Task found when tasklist should be empty'

    # Set up task and getid
    cores = NCORES
    args_for_sim = 'sleep 0'
    task0 = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    taskid = task0.id
    print("taskid is: {}".format(taskid))
    A = exctr.get_task(taskid)
    assert A is task0, 'Task get_task returned unexpected task' + str(A)
    task0 = polling_loop(exctr, task0)

    # Get non-existent taskid
    A = exctr.get_task(taskid+1)
    assert A is None, 'Task found when supplied taskid should not exist'


@pytest.mark.timeout(30)
def test_procs_and_machinefile_logic():
    """ Test of supplying various input configurations."""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))

    # Note: Could test task_partition routine directly - without launching tasks...

    # Testing machinefile
    setup_executor()
    exctr = Executor.executor
    args_for_sim = 'sleep 0'

    machinefilename = "my_machinefile"
    cores = NCORES
    with open(machinefilename, 'w') as f:
        for rank in range(cores):
            f.write(socket.gethostname() + '\n')

    task = exctr.submit(calc_type='sim', machinefile=machinefilename, app_args=args_for_sim)
    task = polling_loop(exctr, task, delay=0.05)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)

    # Testing num_procs = num_nodes*procs_per_node (shouldn't fail)
    task = exctr.submit(calc_type='sim', num_procs=6, num_nodes=2, procs_per_node=3, app_args=args_for_sim)
    task = polling_loop(exctr, task, delay=0.05)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)

    # Testing num_procs not num_nodes*procs_per_node (should fail)
    try:
        task = exctr.submit(calc_type='sim', num_procs=9, num_nodes=2, procs_per_node=5, app_args=args_for_sim)
    except MPIResourcesException as e:
        assert e.args[0] == 'num_procs does not equal num_nodes*procs_per_node'
    else:
        assert 0

    # Testing no num_procs (shouldn't fail)
    task = exctr.submit(calc_type='sim', num_nodes=2, procs_per_node=3, app_args=args_for_sim)
    assert 1
    task = polling_loop(exctr, task, delay=0.05)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)

    # Testing nothing given (should fail)
    try:
        task = exctr.submit(calc_type='sim', app_args=args_for_sim)
    except MPIResourcesException as e:
        assert e.args[0] == 'Need num_procs, num_nodes/procs_per_node, or machinefile'
    else:
        assert 0

    # Testing no num_nodes (shouldn't fail)
    task = exctr.submit(calc_type='sim', num_procs=2, procs_per_node=2, app_args=args_for_sim)
    assert 1
    task = polling_loop(exctr, task, delay=0.05)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)

    # Testing no procs_per_node (shouldn't fail)
    task = exctr.submit(calc_type='sim', num_nodes=1, num_procs=2, app_args=args_for_sim)
    assert 1
    task = polling_loop(exctr, task, delay=0.05)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)


@pytest.mark.timeout(20)
def test_doublekill():
    """Test attempt to kill already killed task

    Kill should have no effect (except warning message) and should remain in state killed
    """
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 2.0'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    task.poll()
    exctr.wait_time = 5

    exctr.kill(task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'USER_KILLED', "task.state should be USER_KILLED. Returned " + str(task.state)
    exctr.kill(task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'USER_KILLED', "task.state should be USER_KILLED. Returned " + str(task.state)


@pytest.mark.timeout(20)
def test_finish_and_kill():
    """Test attempt to kill already finished task

    Kill should have no effect (except warning message) and should remain in state FINISHED
    """
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 0.1'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    while not task.finished:
        time.sleep(0.1)
        task.poll()
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)
    exctr.kill(task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)
    # Try polling after finish - should return with no effect
    task.poll()
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)


@pytest.mark.timeout(20)
def test_launch_and_kill():
    """Test launching and immediately killing tasks with no poll"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 2.0'
    task_list = []
    exctr.wait_time = 1
    for taskid in range(5):
        task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
        exctr.kill(task)
        task_list.append(task)

    for task in task_list:
        assert task.finished, "task.finished should be True. Returned " + str(task.finished)
        assert task.state == 'USER_KILLED', "task.state should be USER_KILLED. Returned " + str(task.state)


def test_launch_as_gen():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 0.1'

    # Try launching as gen when not registered as gen
    try:
        task = exctr.submit(calc_type='gen', num_procs=cores, app_args=args_for_sim)
    except ExecutorException as e:
        assert e.args[0] == 'Default gen app is not set'
    else:
        assert 0

    exctr.register_app(full_path=sim_app, calc_type='gen')
    task = exctr.submit(calc_type='gen', num_procs=cores, app_args=args_for_sim)
    task = polling_loop(exctr, task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)

    # Try launching as 'alloc' which is not a type
    try:
        task = exctr.submit(calc_type='alloc', num_procs=cores, app_args=args_for_sim)
    except ExecutorException as e:
        assert e.args[0] + e.args[1] == 'Unrecognized calculation type' + 'alloc'
    else:
        assert 0


def test_launch_no_app():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor_noapp()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 0.1'
    try:
        _ = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    except ExecutorException as e:
        assert e.args[0] == 'Default sim app is not set'
    else:
        assert 0
    try:
        _ = exctr.submit(num_procs=cores, app_args=args_for_sim)
    except ExecutorException as e:
        assert e.args[0] == 'Either app_name or calc_type must be set'
    else:
        assert 0


def test_kill_task_with_no_submit():
    from libensemble.executors.executor import Task
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor

    # Try kill invalid task
    try:
        exctr.kill('mytask')
    except ExecutorException as e:
        assert e.args[0] == 'Invalid task has been provided'
    else:
        assert 0

    # Create a task directly with no submit (Not supported for users)
    # Debatably make taskID 0 as executor should be deleted if use setup function.
    # But this allows any task ID.
    exp_msg = ('Attempting to kill task libe_task_my_simtask.x_.+that has '
               'no process ID - check tasks been launched')
    exp_re = re.compile(exp_msg)
    myapp = exctr.sim_default_app
    task1 = Task(app=myapp, stdout='stdout.txt')
    try:
        exctr.kill(task1)
    except ExecutorException as e:
        assert bool(re.match(exp_re, e.args[0]))
    else:
        assert 0


def test_poll_task_with_no_submit():
    from libensemble.executors.executor import Task
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor

    # Create a task directly with no submit (Not supported for users)
    exp_msg = ('Polled task libe_task_my_simtask.x_.+ '
               'has no process ID - check tasks been launched')
    exp_re = re.compile(exp_msg)
    myapp = exctr.sim_default_app
    task1 = Task(app=myapp, stdout='stdout.txt')
    try:
        task1.poll()
    except ExecutorException as e:
        assert bool(re.match(exp_re, e.args[0]))
    else:
        assert 0


def test_task_failure():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    cores = NCORES
    args_for_sim = 'sleep 1.0 Fail'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    task = polling_loop(exctr, task)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FAILED', "task.state should be FAILED. Returned " + str(task.state)


def test_retries_launch_fail():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor_fakerunner()
    exctr = Executor.executor
    exctr.retry_delay_incr = 0.05
    cores = NCORES
    args_for_sim = 'sleep 0'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    assert task.state == 'CREATED', "task.state should be CREATED. Returned " + str(task.state)
    assert exctr.mpi_runner.subgroup_launch, "subgroup_launch should be True"
    assert task.run_attempts == 5, "task.run_attempts should be 5. Returned " + str(task.run_attempts)


def test_retries_run_fail():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()
    exctr = Executor.executor
    exctr.retry_delay_incr = 0.05
    cores = NCORES
    args_for_sim = 'sleep 0 Fail'
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim, wait_on_start=True)
    assert task.state == 'FAILED', "task.state should be FAILED. Returned " + str(task.state)
    assert task.run_attempts == 5, "task.run_attempts should be 5. Returned " + str(task.run_attempts)


def test_register_apps():
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    setup_executor()  # This registers an app my_simtask.x (default sim)
    exctr = Executor.executor
    exctr.register_app(full_path='/path/to/fake_app1.x', app_name='fake_app1')
    exctr.register_app(full_path='/path/to/fake_app2.py', app_name='fake_app2')

    # Check selected attributes
    app = exctr.get_app('my_simtask.x')
    assert app.name == 'my_simtask.x'
    assert app.gname == 'libe_app_my_simtask.x'

    app = exctr.get_app('fake_app1')
    assert app.name == 'fake_app1'
    assert app.gname == 'libe_app_fake_app1'
    assert app.exe == 'fake_app1.x'
    assert app.calc_dir == '/path/to'

    app = exctr.get_app('fake_app2')
    assert app.name == 'fake_app2'
    assert app.gname == 'libe_app_fake_app2'

    py_exe, app_exe = app.full_path.split()
    assert os.path.split(py_exe)[1].startswith('python')
    assert app_exe == '/path/to/fake_app2.py'

    try:
        app = exctr.get_app('fake_app3')
    except ExecutorException as e:
        assert e.args[0] == 'Application fake_app3 not found in registry'
        # Ordering of dictionary may vary
        # assert e.args[1] == "Registered applications: ['my_simtask.x', 'fake_app1', 'fake_app2']"


def test_serial_exes():
    setup_serial_executor()
    exctr = Executor.executor
    args_for_sim = 'sleep 0.1'
    task = exctr.submit(calc_type='sim', app_args=args_for_sim, wait_on_start=True)
    task.wait()
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)


def test_serial_startup_times():
    setup_executor_startups()
    exctr = Executor.executor

    t1 = time.time()
    task = exctr.submit(app_name='c_startup')
    task.wait()
    stime = float(task.read_stdout())
    startup_time = stime - t1
    print('start up time for c program', startup_time)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)
    assert 0 < startup_time < 1, "Start up time for C program took " + str(startup_time)

    t1 = time.time()
    task = exctr.submit(app_name='py_startup')
    task.wait()
    stime = float(task.read_stdout())
    startup_time = stime - t1
    print('start up time for python program', startup_time)
    assert task.finished, "task.finished should be True. Returned " + str(task.finished)
    assert task.state == 'FINISHED', "task.state should be FINISHED. Returned " + str(task.state)
    assert 0 < startup_time < 1, "Start up time for python program took " + str(startup_time)


if __name__ == "__main__":
    setup_module(__file__)
    test_launch_and_poll()
    test_launch_and_wait()
    test_launch_and_wait_timeout()
    test_launch_wait_on_start()
    test_kill_on_file()
    test_kill_on_timeout()
    test_kill_on_timeout_polling_loop_method()
    test_launch_and_poll_multitasks()
    test_get_task()
    test_procs_and_machinefile_logic()
    test_doublekill()
    test_finish_and_kill()
    test_launch_and_kill()
    test_launch_as_gen()
    test_launch_no_app()
    test_kill_task_with_no_submit()
    test_poll_task_with_no_submit()
    test_task_failure()
    test_retries_launch_fail()
    test_retries_run_fail()
    test_register_apps()
    test_serial_exes()
    test_serial_startup_times()
    teardown_module(__file__)



__all__ = ['geomedian_eval']
import numpy as np
import numpy.linalg as la


def EvaluateFunction(x, component, B):
    """
    Evaluates the sum of squares-variant chwirut function
    """
    m = B.shape[0]
    assert B.shape[1] == len(x)
    assert 0 <= component <= m-1
    i = component
    b_i = B[i]

    f_i = 1.0/m * la.norm(x-b_i)
    return f_i


def EvaluateJacobian(x, component, B):
    """
    Evaluates the sum of squares-variant chwirut Jacobian
    """
    m = B.shape[0]
    assert B.shape[1] == len(x)
    assert 0 <= component <= m-1
    i = component
    b_i = B[i]

    df_i = 1.0/m * (x-b_i)/la.norm(x-b_i)
    return df_i


def geomedian_eval(H, persis_info, sim_specs, _):

    B = persis_info['params']['B']

    num_xs = len(H['x'])  # b==1 always?
    H_o = np.zeros(num_xs, dtype=sim_specs['out'])

    for k, x in enumerate(H['x']):
        i = H[k]['obj_component']   # f_i

        if H[k]['get_grad']:
            H_o['gradf_i'][k] = EvaluateJacobian(x, i, B)
        else:
            H_o['f_i'][k] = EvaluateFunction(x, i, B)

    return H_o, persis_info



"""
Calls the branin function. Default behavior uses the python function, but
uncommenting lines will write x.in to file, call branin.py, and then read f.out.
"""
import numpy as np
import time
from libensemble.sim_funcs.branin.branin import branin


def call_branin(H, persis_info, sim_specs, _):
    """ Evaluates the Branin function """

    batch = len(H['x'])

    H_o = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        # Uncomment the following if you want to use the file system to do evaluations
        # devnull = open(os.devnull, 'w')
        # np.savetxt('./x.in', x, fmt='%16.16f', delimiter=' ', newline=" ")
        # p = subprocess.call(['python', 'branin.py'], cwd='./', stdout=devnull)
        # H_o['f'][i] = np.loadtxt('./f.out', dtype=float)

        H_o['f'][i] = branin(x[0], x[1])

        if 'user' in sim_specs and 'uniform_random_pause_ub' in sim_specs['user']:
            time.sleep(sim_specs['user']['uniform_random_pause_ub']*np.random.uniform())

    return H_o, persis_info



"""
This module contains functions for implementing an persistent NSGA2 generator
function. The evaluation of the fitness of the current population's members
occurs in `evaluate_pop`, where the points are communicated to the libEnsemble
manager; the manager coordinates their evaluation and then returns their
`fitness_values`.

https://gist.github.com/darden1/fa8f96185a46796ed9516993bfe24862
"""

__all__ = ['deap_nsga2', 'evaluate_pop', 'nsga2_toolbox']

from deap import base, creator, tools
import numpy as np
import array

from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport


def uniform(low, up):
    return [np.random.uniform(a, b) for a, b in zip(low, up)]


def nsga2_toolbox(gen_specs):
    '''
    Returns a DEAP toolbox for use in a NSGA2 loop, derived from `this example.
    <https://github.com/ChristopherMayes/xdeap/blob/master/xdeap/nsga2_tools.py>`_
    '''
    w = gen_specs['user']['weights']
    eta = gen_specs['user']['eta']
    inp = gen_specs['user']['indpb']
    lb = gen_specs['user']['lb']
    ub = gen_specs['user']['ub']

    try:
        del creator.MyFitness
    except Exception:
        pass

    try:
        del creator.Individual
    except Exception:
        pass

    creator.create('MyFitness', base.Fitness, weights=w)
    creator.create('Individual', array.array, typecode='d', fitness=creator.MyFitness)
    toolbox = base.Toolbox()

    toolbox.register('attr_float', uniform, lb, ub)
    toolbox.register('individual', tools.initIterate, creator.Individual, toolbox.attr_float)
    toolbox.register('population', tools.initRepeat, list, toolbox.individual)

    toolbox.register('mate', tools.cxSimulatedBinaryBounded, low=lb, up=ub, eta=eta)
    toolbox.register('mutate', tools.mutPolynomialBounded, low=lb, up=ub, eta=eta, indpb=inp)
    toolbox.register('select', tools.selNSGA2)

    return toolbox


def evaluate_pop(g, deap_object, Out, ps):
    '''
    Evaluates the fitness of a population by communicating the individuals in
    the population to the libEnsemble manager, and then awaiting their fitness_values.
    '''
    # Take population or list of individuals
    # Sending individuals from population to sim to calc fitness
    for index, ind in enumerate(deap_object):
        Out['individual'][index] = ind
        Out['generation'][index] = g
    # Sending work to sim_f, which is defined in main call script
    # A fitness value will be returned in calc_in
    tag, Work, calc_in = ps.send_recv(Out[['individual', 'generation']])

    if tag not in [STOP_TAG, PERSIS_STOP]:
        for i, ind in enumerate(deap_object):
            # Attaching fitness values from sim to population
            # i.e. replacing values with those generated by the sim
            fvals = calc_in['fitness_values'][i]
            ind.fitness.values = [fvals] if isinstance(fvals, float) else list(fvals)

    return deap_object, tag


def deap_nsga2(H, persis_info, gen_specs, libE_info):
    '''
    An implementation of the NSGA2 evolutionary algorithm.
    '''
    # Check to make sure boundaries are list, not array
    assert isinstance(gen_specs['user']['lb'], list), "lb is wrong type, must be a list!"
    assert isinstance(gen_specs['user']['ub'], list), "ub is wrong type, must be a list!"

    # Initialize NSGA2 DEAP toolbox
    toolbox = nsga2_toolbox(gen_specs)
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    pop_size = gen_specs['user']['pop_size']
    # CXPB  is the probability with which two individuals are crossed
    MU, CXPB = pop_size, gen_specs['user']['cxpb']
    pop = toolbox.population(n=MU)  # MU is Population size ( # of individuals)
    Out = np.zeros(pop_size, dtype=gen_specs['out'])

    if len(H):
        tag = None
        g = max(H['generation'])
        individuals = H['individual'][-pop_size:]
        fvals = H['fitness_values'][-pop_size:]
        print("Loading initial collection of points as generation ", g, '.')

        for i, ind in enumerate(pop):
            # Fill in first pop and output with provided points
            ind[:] = array.array('d', individuals[i])
            ind.fitness.values = [fvals[i]] if isinstance(fvals[i], float) else list(fvals[i])
            Out['individual'][i] = individuals[i]
            Out['generation'][i] = g
    else:
        print('No initial sample provided, starting from scratch.')
        g = 0  # generation count
        # Running fitness calc for first generation
        pop, tag = evaluate_pop(g, pop, Out, ps)

    # This is just to assign the crowding distance to the individuals
    # no actual selection is done
    pop = toolbox.select(pop, len(pop))

    # Begin the evolution
    while tag not in [STOP_TAG, PERSIS_STOP]:
        # A new generation
        g = g + 1
        print("-- Generation %i --" % g, flush=True)

        # Apply crossover and mutation on the offspring
        offspring = tools.selTournamentDCD(pop, len(pop))
        offspring = [toolbox.clone(ind) for ind in offspring]

        for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
            if np.random.uniform() <= CXPB:
                toolbox.mate(ind1, ind2)

            toolbox.mutate(ind1)
            toolbox.mutate(ind2)
            del ind1.fitness.values, ind2.fitness.values

        # Evaluate the individuals with an invalid fitness
        # These are individuals who had their fitness deleted by crossover or mutation
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]

        # Need to check that there were invalid in divides first.
        # When using small test number of points (2, 5, etc)
        # There is a probability that there will be no invalid individuals
        if invalid_ind:
            print('Finished evaluating population, doing selection now.')
            # Running fitness calc on gens > 0
            invalid_ind, tag = evaluate_pop(g, invalid_ind, Out, ps)
            if tag not in [STOP_TAG, PERSIS_STOP]:
                # Select the next generation population
                pop = toolbox.select(pop + offspring, MU)
        else:
            print('There were no invalid individuals')
            # Don't update population
            pass

        fits = np.array(np.array([ind.fitness.values for ind in pop]))
        if tag in [STOP_TAG, PERSIS_STOP]:
            # Min value when exiting
            print('Met exit criteria. Current best fitness is:', np.min(fits))
        else:
            print('Current fitness minimum:', np.min(fits, axis=0))
            print('Sum of fit values at end of loop', sum(fits))

    Out['last_points'] = 1
    return Out, persis_info, FINISHED_PERSISTENT_GEN_TAG



"""
Tests the batch-mode of the Tasmanian generator function.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_tasmanian.py
   python3 test_persistent_tasmanian.py --nworkers 3 --comms local
   python3 test_persistent_tasmanian.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local
# TESTSUITE_NPROCS: 4
# TESTSUITE_OS_SKIP: OSX
# TESTSUITE_EXTRA: true

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f
from libensemble.gen_funcs.persistent_tasmanian import sparse_grid_batched as gen_f_batched
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from time import time


def tasmanian_init_global():
    # Note: if Tasmanian has been compiled with OpenMP support (i.e., the usual way)
    #       libEnsemble calls cannot be made after the `import Tasmanian` clause
    #       there is a conflict between the OpenMP environment and Python threading
    #       thus Tasmanian has to be imported inside the `tasmanian_init` method
    import Tasmanian

    grid = Tasmanian.makeGlobalGrid(num_dimensions, 1, 6, "iptotal", "clenshaw-curtis")
    grid.setDomainTransform(np.array([[-5.0, 5.0], [-2.0, 2.0]]))
    return grid


def tasmanian_init_localp():
    import Tasmanian

    grid = Tasmanian.makeLocalPolynomialGrid(num_dimensions, 1, 3)
    grid.setDomainTransform(np.array([[-5.0, 5.0], [-2.0, 2.0]]))
    return grid


nworkers, is_manager, libE_specs, _ = parse_args()

if is_manager:
    start_time = time()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

num_dimensions = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f_batched,
    'persis_in': ['x', 'f', 'sim_id'],
    'out': [('x', float, num_dimensions)],
}

alloc_specs = {'alloc_f': alloc_f}

grid_files = []

for run in range(3):
    # testing two cases, static construction without refinement
    # and refinement until 100 points have been computed
    persis_info = add_unique_random_streams({}, nworkers + 1)

    # set the stopping criteria
    if run != 1:
        # note that using 'setAnisotropicRefinement' without 'gen_max' will create an infinite loop
        # other stopping criteria could be used with 'setSurplusRefinement' or no refinement
        exit_criteria = {'elapsed_wallclock_time': 10}
    elif run == 1:
        exit_criteria = {'gen_max': 100}  # This will test persistent_tasmanian stopping early.

    # tasmanian_init has to be a method that returns an initialized TasmanianSparseGrid object
    # tasmanian_checkpoint_file will be overwritten between each step of the iterative refinement
    #   the final grid will also be stored in the file
    gen_specs['user'] = {
        'tasmanian_init': tasmanian_init_global if run < 2 else tasmanian_init_localp,
        'tasmanian_checkpoint_file': 'tasmanian{0}.grid'.format(run),
    }

    # setup the refinement criteria
    if run == 0:
        gen_specs['user']['refinement'] = 'none'

    if run == 1:
        # See Tasmanian manual: https://ornl.github.io/TASMANIAN/stable/classTasGrid_1_1TasmanianSparseGrid.html
        gen_specs['user']['refinement'] = 'setAnisotropicRefinement'
        gen_specs['user']['sType'] = 'iptotal'
        gen_specs['user']['iMinGrowth'] = 10
        gen_specs['user']['iOutput'] = 0

    if run == 2:
        gen_specs['user']['refinement'] = 'setSurplusRefinement'
        gen_specs['user']['fTolerance'] = 1.0e-2
        gen_specs['user']['sCriteria'] = 'classic'
        gen_specs['user']['iOutput'] = 0

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

    if is_manager:
        grid_files.append(gen_specs['user']['tasmanian_checkpoint_file'])

        if run == 0:
            print('[Manager]: Time taken =', time() - start_time, flush=True)

            save_libE_output(H, persis_info, __file__, nworkers)

if is_manager:
    # run sanity check on the computed results
    # using the checkpoint file to read the grids from the filesystem
    # Note: cannot make any more libEnsemble calls after importing Tasmanian,
    #       see the earlier note in tasmanian_init_global()
    import Tasmanian

    assert len(grid_files) == 3, "Failed to generate three Tasmanian grid files"

    for run in range(len(grid_files)):
        grid = Tasmanian.SparseGrid()
        grid.read(grid_files[run])
        if run == 0:
            assert grid.getNumNeeded() == 0, "Failed to leave no points needing data"
            assert grid.getNumLoaded() == 49, "Failed to load all points"

        if run == 1:
            assert grid.getNumNeeded() == 0, "Failed to stop after completing the refinement iteration"
            assert grid.getNumLoaded() == 89, "Failed to load all points"

        if run == 2:
            assert grid.getNumNeeded() == 0, "Failed to stop after completing the refinement iteration"
            assert grid.getNumLoaded() == 93, "Failed to load all points"



import numpy as np
import copy

branin_vals_and_minima = np.array(
    [
        [-3.14159, 12.275, 0.397887],
        [3.14159, 2.275, 0.397887],
        [9.42478, 2.475, 0.397887],
    ]
)

six_hump_camel_minima = np.array(
    [
        [-0.089842, 0.712656],
        [0.089842, -0.712656],
        [-1.70361, 0.796084],
        [1.70361, -0.796084],
        [-1.6071, -0.568651],
        [1.6071, 0.568651],
    ]
)


def nan_func(calc_in, persis_info, sim_specs, libE_info):
    H = np.zeros(1, dtype=sim_specs['out'])
    H['f_i'] = np.nan
    H['f'] = np.nan
    return (H, persis_info)


def write_sim_func(calc_in, persis_info, sim_specs, libE_info):
    out = np.zeros(1, dtype=sim_specs['out'])
    out['f'] = calc_in['x']
    with open('test_sim_out.txt', 'a') as f:
        f.write('sim_f received: {}\n'.format(out['f']))
    return out, persis_info


def write_uniform_gen_func(H, persis_info, gen_specs, _):
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(lb)
    b = gen_specs['user']['gen_batch_size']
    H_o = np.zeros(b, dtype=gen_specs['out'])
    H_o['x'] = persis_info['rand_stream'].uniform(lb, ub, (b, n))
    with open('test_gen_out.txt', 'a') as f:
        f.write('gen_f produced: {}\n'.format(H_o['x']))
    return H_o, persis_info


uniform_or_localopt_gen_out = [
    ('priority', float),
    ('local_pt', bool),
    ('known_to_aposmm', bool),
    ('dist_to_unit_bounds', float),
    ('dist_to_better_l', float),
    ('dist_to_better_s', float),
    ('ind_of_better_l', int),
    ('ind_of_better_s', int),
    ('started_run', bool),
    ('num_active_runs', int),
    ('local_min', bool),
]

aposmm_gen_out = copy.deepcopy(uniform_or_localopt_gen_out)
aposmm_gen_out += [
    ('sim_id', int),
    ('paused', bool),
    ('pt_id', int),  # Identify the same point evaluated by different sim_f's or components
]

# give_sim_work_first persis_info
persis_info_1 = {
    'total_gen_calls': 0,  # Counts gen calls in alloc_f
    'last_worker': 0,  # Remembers last gen worker in alloc_f
    'next_to_give': 0,  # Remembers next H row to give in alloc_f
}

persis_info_1[0] = {
    'run_order': {},  # Used by manager to remember run order
    'old_runs': {},  # Used by manager to store old runs order
    'total_runs': 0,  # Used by manager to count total runs
    'rand_stream': np.random.RandomState(1),
}
# end_persis_info_rst_tag

persis_info_2 = copy.deepcopy(persis_info_1)
persis_info_2[1] = persis_info_2[0]
persis_info_2.pop(0)

# give_sim_work_first_pausing persis_info
persis_info_3 = copy.deepcopy(persis_info_1)
persis_info_3.pop('next_to_give')
persis_info_3['need_to_give'] = set()
persis_info_3['complete'] = set()
persis_info_3['has_nan'] = set()
persis_info_3['already_paused'] = set()
persis_info_3['H_len'] = 0
persis_info_3['best_complete_val'] = np.inf
persis_info_3['local_pt_ids'] = set()
persis_info_3['inds_of_pt_ids'] = {}



#!/usr/bin/env python
import os
import numpy as np
from forces_simf import run_forces  # Sim func from current dir

# Import libEnsemble modules
from libensemble.libE import libE
from libensemble.manager import ManagerException
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble import logger
from forces_support import test_libe_stats, test_ensemble_dir, check_log_exception

USE_BALSAM = False
PERSIS_GEN = False

if PERSIS_GEN:
    from libensemble.gen_funcs.persistent_uniform_sampling import persistent_uniform as gen_f
    from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
else:
    from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
    from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first as alloc_f


logger.set_level('INFO')  # INFO is now default

nworkers, is_manager, libE_specs, _ = parse_args()

if is_manager:
    print('\nRunning with {} workers\n'.format(nworkers))

sim_app = os.path.join(os.getcwd(), 'forces.x')

# Normally would be pre-compiled
if not os.path.isfile('forces.x'):
    if os.path.isfile('build_forces.sh'):
        import subprocess
        subprocess.check_call(['./build_forces.sh'])

# Create executor and register sim to it.
if USE_BALSAM:
    from libensemble.executors.balsam_executor import BalsamMPIExecutor
    exctr = BalsamMPIExecutor()
else:
    from libensemble.executors.mpi_executor import MPIExecutor
    exctr = MPIExecutor()
exctr.register_app(full_path=sim_app, app_name='forces')

# Note: Attributes such as kill_rate are to control forces tests, this would not be a typical parameter.

# State the objective function, its arguments, output, and necessary parameters (and their sizes)
sim_specs = {'sim_f': run_forces,         # Function whose output is being minimized
             'in': ['x'],                 # Name of input for sim_f
             'out': [('energy', float)],  # Name, type of output from sim_f
             'user': {'keys': ['seed'],
                      'cores': 2,
                      'sim_particles': 1e3,
                      'sim_timesteps': 5,
                      'sim_kill_minutes': 10.0,
                      'particle_variance': 0.2,
                      'kill_rate': 0.5,
                      'fail_on_sim': False,
                      'fail_on_submit': False}  # Won't occur if 'fail_on_sim' True
             }
# end_sim_specs_rst_tag

# State the generating function, its arguments, output, and necessary parameters.
gen_specs = {'gen_f': gen_f,                  # Generator function
             'in': [],                        # Generator input
             'out': [('x', float, (1,))],     # Name, type and size of data produced (must match sim_specs 'in')
             'user': {'lb': np.array([0]),             # Lower bound for random sample array (1D)
                      'ub': np.array([32767]),         # Upper bound for random sample array (1D)
                      'gen_batch_size': 1000,          # How many random samples to generate in one call
                      }
             }

if PERSIS_GEN:
    alloc_specs = {'alloc_f': alloc_f}
else:
    alloc_specs = {'alloc_f': alloc_f,
                   'user': {'batch_mode': True,    # If true wait for all sims to process before generate more
                            'num_active_gens': 1}  # Only one active generator at a time
                   }

libE_specs['save_every_k_gens'] = 1000  # Save every K steps
libE_specs['sim_dirs_make'] = True      # Separate each sim into a separate directory
libE_specs['profile'] = False    # Whether to have libE profile on (default False)

# Maximum number of simulations
sim_max = 8
exit_criteria = {'sim_max': sim_max}

# Create a different random number stream for each worker and the manager
persis_info = {}
persis_info = add_unique_random_streams(persis_info, nworkers + 1)

try:
    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria,
                                persis_info=persis_info,
                                alloc_specs=alloc_specs,
                                libE_specs=libE_specs)

except ManagerException:
    if is_manager and sim_specs['user']['fail_on_sim']:
        check_log_exception()
        test_libe_stats('Exception occurred\n')
else:
    if is_manager:
        save_libE_output(H, persis_info, __file__, nworkers)
        if sim_specs['user']['fail_on_submit']:
            test_libe_stats('Task Failed\n')
        test_ensemble_dir(libE_specs, './ensemble', nworkers, sim_max)



import os
import shutil
import time

from libensemble.executors.executor import Task, Executor, ExecutorException
from libensemble.executors.mpi_executor import MPIExecutor


def setup_module(module):
    print("setup_module      module:%s" % module.__name__)
    if Executor.executor is not None:
        del Executor.executor
        Executor.executor = None


def setup_function(function):
    print("setup_function    function:%s" % function.__name__)
    if Executor.executor is not None:
        del Executor.executor
        Executor.executor = None


def teardown_module(module):
    print("teardown_module   module:%s" % module.__name__)
    if Executor.executor is not None:
        del Executor.executor
        Executor.executor = None


def test_task_funcs():
    dummyappname = os.getcwd() + '/myapp.x'
    exctr = MPIExecutor()
    exctr.register_app(full_path=dummyappname, calc_type='gen', desc='A dummy calc')
    exctr.register_app(full_path=dummyappname, calc_type='sim', desc='A dummy calc')

    dirname = 'dir_taskc_tests'
    if os.path.exists(dirname):
        shutil.rmtree(dirname)
    os.mkdir(dirname)
    os.chdir(dirname)
    myworkdir = os.getcwd()

    # First try no app - check exception raised?
    jc_triggered = False
    try:
        _ = Task(workdir=myworkdir, stdout='stdout.txt', stderr='stderr.txt')
    except ExecutorException:
        jc_triggered = True

    assert jc_triggered, "Failed to raise exception if create task with no app"

    # Now with no workdir specified
    dummyapp = exctr.gen_default_app
    task1 = Task(app=dummyapp, stdout='stdout.txt', stderr='stderr.txt')
    wd_exist = task1.workdir_exists()
    assert not wd_exist  # , "No workdir specified, yet workdir_exists does not return False"
    stdout_exist = task1.stdout_exists()
    assert not stdout_exist
    f_exist = task1.file_exists_in_workdir('running_output.txt')
    assert not f_exist

    # Create task properly specified
    task2 = Task(app=dummyapp, workdir=myworkdir, stdout='stdout.txt', stderr='stderr.txt')

    # Workdir does exist
    wd_exist = task2.workdir_exists()
    assert wd_exist

    # Files do not exist
    stdout_exist = task2.stdout_exists()
    assert not stdout_exist
    stderr_exist = task2.stderr_exists()
    assert not stderr_exist
    f_exist = task2.file_exists_in_workdir('running_output.txt')
    assert not f_exist

    valerr_triggered = False
    try:
        task2.read_stdout()
    except ValueError:
        valerr_triggered = True
    assert valerr_triggered

    valerr_triggered = False
    try:
        task2.read_file_in_workdir('running_output.txt')
    except ValueError:
        valerr_triggered = True
    assert valerr_triggered

    # Now create files and check positive results
    with open("stdout.txt", "w") as f:
        f.write('This is stdout')
    with open("stderr.txt", "w") as f:
        f.write('This is stderr')
    with open("running_output.txt", "w") as f:
        f.write('This is running output')

    # task2 = Task(app = dummyapp, workdir = myworkdir, stdout = 'stdout.txt')
    # wd_exist = task2.workdir_exists()
    # assert wd_exist
    stdout_exist = task2.stdout_exists()
    assert stdout_exist
    stderr_exist = task2.stderr_exists()
    assert stderr_exist
    f_exist = task2.file_exists_in_workdir('running_output.txt')
    assert f_exist
    assert 'This is stdout' in task2.read_stdout()
    assert 'This is stderr' in task2.read_stderr()
    assert 'This is running output' in task2.read_file_in_workdir('running_output.txt')

    # Check if workdir does not exist
    task2.workdir = task2.workdir + '/bubbles'
    wd_exist = task2.workdir_exists()
    assert not wd_exist

    # Check timing
    assert not task2.submit_time and not task2.runtime and not task2.total_time
    task2.calc_task_timing()
    assert not task2.submit_time and not task2.runtime and not task2.total_time
    task2.submit_time = time.time()
    task2.calc_task_timing()
    assert task2.runtime is not None and task2.runtime == task2.total_time
    save_runtime, save_total_time = task2.runtime, task2.total_time
    task2.calc_task_timing()
    assert save_runtime == task2.runtime
    assert save_total_time == task2.total_time

    # Clean up
    os.chdir('../')
    shutil.rmtree(dirname)


if __name__ == "__main__":
    test_task_funcs()



"""
This module for detects and returns intranode resources

"""

import os
import psutil
import logging
import collections

logger = logging.getLogger(__name__)

REMOTE_LAUNCH_LIST = ['aprun', 'jsrun', 'srun']  # Move to feature of mpi_runner


def get_cpu_cores(hyperthreads=False):
    """Returns the number of cores on the node.

    If hyperthreads is true, this is the logical CPU cores; else
    the physical cores are returned.

    Note: This returns cores available on the current node. It will
    not work for systems of multiple node types
    """
    return psutil.cpu_count(logical=hyperthreads)  # This is ranks available per node


def _get_local_cpu_resources():
    """Returns logical and physical cores on the local node"""
    logical_cores_avail_per_node = get_cpu_cores(hyperthreads=True)
    physical_cores_avail_per_node = get_cpu_cores(hyperthreads=False)
    return (physical_cores_avail_per_node, logical_cores_avail_per_node)


def _print_local_cpu_resources():
    """Prints logical and physical cores on the local node"""
    cores_info = _get_local_cpu_resources()
    print(cores_info[0], cores_info[1], flush=True)


def _get_remote_cpu_resources(launcher):
    """Launches a probe job to obtain logical and physical cores on remote node"""
    import subprocess
    output = subprocess.check_output([launcher, 'python', __file__])
    return output.decode()


def _get_cpu_resources_from_env(env_resources=None):
    """Returns logical and physical cores per node by querying environment or None"""

    if not env_resources:
        return None

    found_count = False
    if env_resources.nodelists['LSF'] in os.environ:
        full_list = os.environ.get(env_resources.nodelists['LSF']).split()
        nodes = [n for n in full_list if 'batch' not in n]
        counter = list(collections.Counter(nodes).values())
        found_count = True
    elif env_resources.nodelists['LSF_shortform'] in os.environ:
        full_list = os.environ.get(env_resources.nodelists['LSF_shortform']).split()
        iter_list = iter(full_list)
        zipped_list = list(zip(iter_list, iter_list))
        nodes_with_count = [n for n in zipped_list if 'batch' not in n[0]]
        counter = [int(n[1]) for n in nodes_with_count]
        found_count = True

    if found_count:
        # Check all nodes have equal cores -  Not doing for other methods currently.
        if len(set(counter)) != 1:
            logger.warning("Detected compute nodes have different core counts: {}".format(set(counter)))

        physical_cores_avail_per_node = counter[0]
        logical_cores_avail_per_node = counter[0]  # How to get SMT threads remotely
        logger.warning("SMT currently not detected, returning physical cores only. Specify procs_per_node to override")
        return (physical_cores_avail_per_node, logical_cores_avail_per_node)
    else:
        return None


def get_sub_node_resources(launcher=None, remote_mode=False, env_resources=None):
    """Returns logical and physical cores per node as a tuple"""
    remote_detection = False
    if remote_mode:
        # May be unnecessary condition
        if launcher in REMOTE_LAUNCH_LIST:
            cores_info = _get_cpu_resources_from_env(env_resources=env_resources)
            if cores_info:
                return (cores_info)
            remote_detection = True  # Cannot obtain from environment

    if remote_detection:
        cores_info_str = _get_remote_cpu_resources(launcher=launcher)
        cores_log, cores_phy, *_ = cores_info_str.split()
        cores_info = (int(cores_log), int(cores_phy))
    else:
        cores_info = _get_local_cpu_resources()
    return (cores_info)


if __name__ == "__main__":
    _print_local_cpu_resources()



"""
Test the APOSMM generator function's capabilities to properly exit when a
timeout has occurred.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_timeout.py
   python3 test_persistent_aposmm_timeout.py --nworkers 3 --comms local
   python3 test_persistent_aposmm_timeout.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi tcp
# TESTSUITE_NPROCS: 4
# TESTSUITE_EXTRA: true

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.periodic_func import func_wrapper as sim_f
from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f
from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams, save_libE_output

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_out = [
    ('x', float, n),
    ('x_on_cube', float, n),
    ('sim_id', int),
    ('local_min', bool),
    ('local_pt', bool),
]

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'localopt_method': 'LN_BOBYQA',
        'xtol_abs': 1e-8,
        'ftol_abs': 1e-8,
        'run_max_eval': 30,
        'lb': np.array([0, -np.pi / 2]),
        'ub': np.array([2 * np.pi, 3 * np.pi / 2]),
        'periodic': True,
        'print': True,
    },
}

alloc_specs = {'alloc_f': alloc_f}

# Setting a very high sim_max value and a short elapsed_wallclock_time so timeout will occur
exit_criteria = {'sim_max': 50000, 'elapsed_wallclock_time': 5}

persis_info = add_unique_random_streams({}, nworkers + 1)

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert flag == 2, "Test should have timed out"
    assert persis_info[1].get('run_order'), "Run_order should have been given back"
    min_ids = np.where(H['local_min'])
    save_libE_output(H, persis_info, __file__, nworkers)



"""
TCP-based bidirectional communicator
------------------------------------

"""

from libensemble.comms.comms import QComm
from multiprocessing.managers import BaseManager
from multiprocessing import Queue


class ServerQCommManager:
    """Set up a QComm manager server.

    The QComm manager server provides shared (networked) access to message
    queues for communication between the libensemble manager and workers.
    """

    def __init__(self, port, authkey):
        """Initialize the server on localhost at an indicated TCP port and key."""
        queues = {'shared': Queue()}

        class ServerQueueManager(BaseManager):
            pass

        def get_queue(name):
            if name not in queues:
                queues[name] = Queue()
            return queues[name]

        ServerQueueManager.register('get_queue', callable=get_queue)
        self.manager = ServerQueueManager(address=('', port), authkey=authkey)
        self.manager.start()

    def shutdown(self):
        """Shutdown the manager"""
        self.manager.shutdown()

    @property
    def address(self):
        """Get IP address for socket."""
        return self.manager.address

    def get_queue(self, name):
        """Get a queue from the shared manager"""
        return self.manager.get_queue(name)

    def get_inbox(self, workerID):
        """Get a worker inbox queue."""
        return self.get_queue('inbox{}'.format(workerID))

    def get_outbox(self, workerID):
        """Get a worker outbox queue."""
        return self.get_queue('outbox{}'.format(workerID))

    def get_shared(self):
        """Get a shared queue for worker subscription."""
        return self.get_queue('shared')

    def await_workers(self, nworkers):
        """Wait for a pool of workers to join."""
        sharedq = self.get_shared()
        wqueues = []
        for _ in range(nworkers):
            workerID = sharedq.get()
            inbox = self.get_outbox(workerID)
            outbox = self.get_inbox(workerID)
            wqueues.append(QComm(inbox, outbox))
        return wqueues

    def __enter__(self):
        """Context enter."""
        return self

    def __exit__(self, etype, value, traceback):
        """Context exit."""
        self.shutdown()


class ClientQCommManager:
    """Set up a client to the QComm server.

    The client runs at the worker and mediates access to the shared queues
    provided by the server.
    """

    def __init__(self, ip, port, authkey, workerID):
        """Attach by TCP to (ip, port) with a uniquely given workerID"""
        self.workerID = workerID

        class ClientQueueManager(BaseManager):
            pass

        ClientQueueManager.register('get_queue')
        self.manager = ClientQueueManager(address=(ip, port), authkey=authkey)
        self.manager.connect()
        sharedq = self.get_shared()
        sharedq.put(workerID)

    def get_queue(self, name):
        """Get a queue from the server."""
        return self.manager.get_queue(name)

    def get_inbox(self):
        """Get this worker's inbox."""
        return self.get_queue('inbox{}'.format(self.workerID))

    def get_outbox(self):
        """Get this worker's outbox."""
        return self.get_queue('outbox{}'.format(self.workerID))

    def get_shared(self):
        """Get the shared queue for worker sign-up."""
        return self.get_queue('shared')

    def __enter__(self):
        """Enter the context."""
        return QComm(self.get_inbox(), self.get_outbox())

    def __exit__(self, etype, value, traceback):
        """Exit the context."""
        pass



"""
Example of multi-fidelity optimization using a persistent GP gen_func (calling
dragonfly) and an algebraic sim_f (that doesn't change with the amount of
resources give).

Execute via one of the following commands (e.g. 5 workers):
   mpiexec -np 5 python3 test_persistent_gp.py
   python3 test_persistent_gp.py --nworkers 4 --comms local
   python3 test_persistent_gp.py --nworkers 4 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 3, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi
# TESTSUITE_NPROCS: 5
# TESTSUITE_EXTRA: true
# TESTSUITE_OS_SKIP: OSX

import numpy as np
from libensemble.libE import libE
from libensemble import logger
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens
from libensemble.tools import add_unique_random_streams
from libensemble.tools import parse_args
from libensemble.message_numbers import WORKER_DONE
from libensemble.gen_funcs.persistent_gp import (persistent_gp_gen_f,
                                                 persistent_gp_mf_gen_f,
                                                 persistent_gp_mf_disc_gen_f)

import warnings

# Dragonfly uses a deprecated np.asscalar command.
warnings.filterwarnings("ignore", category=DeprecationWarning)

nworkers, is_manager, libE_specs, _ = parse_args()


def run_simulation(H, persis_info, sim_specs, libE_info):
    # Extract input parameters
    values = list(H['x'][0])
    x0 = values[0]
    x1 = values[1]
    # Extract fidelity parameter
    z = H['z'][0]

    libE_output = np.zeros(1, dtype=sim_specs['out'])
    calc_status = WORKER_DONE

    # Function that depends on the resolution parameter
    libE_output['f'] = -(x0 + 10 * np.cos(x0 + 0.1 * z)) * (x1 + 5 * np.cos(x1 - 0.2 * z))

    return libE_output, persis_info, calc_status


sim_specs = {
    'sim_f': run_simulation,
    'in': ['x', 'z'],
    'out': [('f', float)],
}

gen_specs = {
    # Generator function. Will randomly generate new sim inputs 'x'.
    'gen_f': persistent_gp_gen_f,
    # Generator input. This is a RNG, no need for inputs.
    'persis_in': ['sim_id', 'x', 'f', 'z'],
    'out': [
        # parameters to input into the simulation.
        ('x', float, (2,)),
        ('z', float),
        ('resource_sets', int)
    ],
    'user': {
        'range': [1, 8],
        'cost_func': lambda z: z[0],
        # Total max number of sims running concurrently.
        'gen_batch_size': nworkers - 1,
        # Lower bound for the n parameters.
        'lb': np.array([0, 0]),
        # Upper bound for the n parameters.
        'ub': np.array([15, 15]),
    },
}

alloc_specs = {
    'alloc_f': only_persistent_gens,
    'user': {'async_return': True},
}

# libE logger
logger.set_level('INFO')

# Exit criteria
exit_criteria = {'sim_max': 10}  # Exit after running sim_max simulations

persis_info = add_unique_random_streams({}, nworkers + 1)

# Run LibEnsemble, and store results in history array H
for run in range(3):
    # Create a different random number stream for each worker and the manager
    persis_info = add_unique_random_streams({}, nworkers + 1)

    if run == 1:
        gen_specs['gen_f'] = persistent_gp_mf_gen_f

    elif run == 2:
        gen_specs['gen_f'] = persistent_gp_mf_disc_gen_f
        gen_specs['user']['cost_func'] = lambda z: z[0][0]**3

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

    if is_manager:
        if run == 0:
            assert not len(np.unique(H['resource_sets'])) > 1, \
                "Resource sets should be the same"

        else:
            assert len(np.unique(H['resource_sets'])) > 1, \
                "Resource sets should be variable."



from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def ensure_one_active_gen(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function gives (in order) entries in ``H`` to idle workers
    to evaluate in the simulation function. The fields in ``sim_specs['in']``
    are given. If there is no active generator, then one is started.

    .. seealso::
        `test_fast_alloc.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_fast_alloc.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']
    support = AllocSupport(W, manage_resources, persis_info, sched_opts)

    Work = {}
    gen_flag = True
    gen_in = gen_specs.get('in', [])

    for wid in support.avail_worker_ids():

        # Skip any cancelled points
        while persis_info['next_to_give'] < len(H) and H[persis_info['next_to_give']]['cancel_requested']:
            persis_info['next_to_give'] += 1

        if persis_info['next_to_give'] < len(H):
            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], [persis_info['next_to_give']], [])
            except InsufficientFreeResources:
                break
            persis_info['next_to_give'] += 1

        elif not support.test_any_gen() and gen_flag:

            if not support.all_returned(H):
                break

            # Give gen work
            return_rows = range(len(H)) if gen_in else []
            try:
                Work[wid] = support.gen_work(wid, gen_in, return_rows, persis_info.get(wid))
            except InsufficientFreeResources:
                break
            gen_flag = False
            persis_info['total_gen_calls'] += 1

    return Work, persis_info



"""
Wrapper for the VTMOP genfuncs drivers, for solving multiobjective
optimization problems with arbitrary number of objectives.
"""
import numpy as np
import ctypes
from sys import float_info
from libensemble.gen_funcs.vtmop_libe import vtmop_libe_mod


def vtmop_gen(H, persis_info, gen_specs, _):
    """
     This generator function solves multiobjective optimization problems
     with d design variables (subject to simple bound constraints) and
     p objectives using VTMOP.

     This generator produces batches of simulation inputs to evaluate, using
     the Fortran package VTMOP. In order to facilitate an interface between
     Fortran and Python, wrappers have been generated for a subset of VTMOP's
     functionality using fmodpy.

     The first time that this generator is called on your machine, the
     necessary VTMOP libraries will be automatically built using gfortran.
     If a different compiler is desired, edit lines 16 and 20 of
     gen_funcs/vtmop_libe/vtmop_libe_python_wrapper.py to specify your
     compiler and the appropriate compiler flags, respectively.

     After the Fortran and C binaries have been built, VTMOP's Python wrapper
     will be automatically imported every time the generator is called without
     need to recompile. Note that this wrapper uses a .so file, which is only
     supported by Mac and Linux/Unix systems (will not work on Windows).

     All inputs and arguments to VTMOP are specified through the
     gen_specs['user'] dictionary.

     Required args:

         gen_specs['user']['d'] (int): is the dimension of the design space.

         gen_specs['user']['p'] (int): is the number of objectives.

         gen_specs['user']['lb'] (np.ndarray): is a 1d array of length D,
             containing the lower bound constraints on the design variables.

         gen_specs['user']['ub'] (np.ndarray): is a 1d array of length D,
             containing the upper bound constraints on the design variables.
             Note that gen_specs['user']['ub'] must be elementwise greater than
             gen_specs['user']['lb'] (up to the design tolerance, specified
             below).

         gen_specs['user']['new_run'] (bool): should be True if you are
             starting a new run, and False if you would like to reload an old
             run using the checkpoint file 'vtmop.chkpt' in your current
             working directory.

         gen_specs['user']['isnb'] (int): is the preferred initial search batch
             size and is used as the size of the Latin hypercube design in the
             zeroeth iteration (initial search). ISNB should be much greater
             than SNB and a multiple of the number of simulation workers.
             Empirically, a good starting place is
             gen_specs['user']['isnb'] ~ 1000. If a large initial database
             is given, it is possible to set gen_specs['user']['isnb'] = 0.

         gen_specs['user']['snb'] (int): is the preferred search batch size
             for all other search steps and is used as the size of the Latin
             hypercube design in the search phase. Empirically, a good starting
             place is gen_specs['user']['snb'] ~ 2 * d.

         gen_specs['user']['onb'] (int): is the preferred batch size for all
             optimization phases. VTMOP will pad out batches of candidates in
             order to produce a multiple of gen_specs['user']['onb'] candidates
             in each iteration. In almost every case, gen_specs['user']['onb']
             should be equal to the number of available simulation workers.

     Optional args:

         gen_specs['user']['lopt_budget'] (int): specifies the budget for GPS
             optimizer iteration when solving the surrogate problems. The
             default value is LOPT_BUDGET is 2500 surrogate evaluations.

         gen_specs['user']['decay'] (float): specifies the decay rate for the
             local trust region (LTR) radius. This value affects how many
             times an isolated point can be the center of a LTR before it i
             discarded. By default, DECAY = 0.5.

         gen_specs['user']['des_tol'] (float): the tolerance for the design
             space. A design point that is within DES_TOL of another evaluated
             design point will not be reevaluated. The default value for
             DES_TOL is the square-root of the working precision EPS (below).
             Note that any value that is smaller than the working precsion EPS
             will be ignored and EPS will be used.

         gen_specs['user']['eps'] (float): specifies the working precision of the
             machine. The default value for EPS is SQRT(EPSILON), where EPSILON
             is the unit roundoff. Note that if the value supplied is smaller
             than the default value then the default value will be used.

         gen_specs['user']['epsw'] (float): is a small positive number, which is
             used as the fudge factor for zero-valued weights. A zero-valued
             weight does not guarantee Pareto optimality. Therefore, all zero
             weights are set to EPSW. The appropriate value of EPSW is problem
             dependent. By default, EPSW is the fourth root of EPSILON (the
             unit roundoff). Note that any value that is smaller than
             SQRT(EPSILON) is ignored and SQRT(EPSILON) will be used.

         gen_specs['user']['obj_tol'] (float): is the tolerance for the
             objective space. An objective point that is within OBJ_TOL of
             being dominated by another objective point will be treated as such.
             The default value of OBJ_TOL is the square-root of EPS. This
             value should be strictly greater than the value of EPS. Note that
             any value that is smaller than the working precsion EPS will be
             ignored and EPS will be used.

         gen_specs['user']['obj_bounds'] (np.ndarray) is an optional 2d array
             of dimensions p by 2, whose first column is a list of lower
             bounds and whose second column is a list of upper bounds on the
             range of interesting objective values. When present, this value
             is used to prune the list of potential LTR centers, so
             that only objective values in the specified range are used when
             looking for gaps in the Pareto. In particular, an objective value
             F(x) will be considered for a LTR center if and only if
             OBJ_BOUNDS(I,1) .LE. F_I(x) .LE. OBJ_BOUNDS(I,2) for all
             I = 1, ..., P. By default, there are no bounds on the interesting
             range. Note that this value is intentionally not saved during
             checkpointing, and must be reset by the user when recovery mode is
             active, whenever a non-default value is desired. This is the only
             input, for which changing the value after loading from a previous
             checkpoint is not ill-advised.

         gen_specs['user']['min_radf'] (float): is the smallest value for the
             fraction r defining the trust region box dimensions r * (UB - LB),
             before an isolated point is abandoned. By default,
             MIN_RADF = 0.1 * TRUST_RADF, and is also set to this default
             value if it is less than DES_TOL. After MIN_RADF and TRUST_RADF
             are set, MIN_RADF < TRUST_RADF must hold.

         gen_specs['user']['trust_radf'] (float): defines the initial trust
             region centered at an isolated point X as
             [X - TRUST_RADF * (UB - LB), X + TRUST_RADF * (UB - LB)]
             intersected with [LB, UB].  By default, TRUST_RADF = 0.2, and is
             also set to this value if the value given is outside the interval
             (DES_TOL, 1 - DES_TOL).

         gen_specs['user']['pmode'] (bool): specifies whether or not iteration
             tasks should be performed in parallel. By default, PMODE = False.
             If changed to True, VTMOP uses OpenMP parallelism to perform
             iteration tasks, and uses the environment variable OMP_NUM_THREADS
             to determine the number of parallel threads per generation.

    Function values/simulation database; VTMOP can also be initialized with
    a precomputed database of n pre-evaluated simulations. To do so, initialize
    a history array H0 as follows:

         H0['x'] (np.ndarray): is a 2d n-by-d history array containing a list
             of precomputed design points.

         H0['f'] (np.ndarray): is a 2d n-by-p history array containing the list
             of corresponding precomputed objective values.

         H0['sim_id'] (np.ndarray): Set H0['sim_id'] = range(n).

         H0[['given', 'returned']] = True

     A checkpoint file (``vtmop.chkpt``) will be generated in the calling
     directory to save the status of VTMOP, and can be used to recover in
     the event of a crash, or can be used to reload any run after pausing.

    """

    # First get the problem dimensions and data
    d = np.int32(gen_specs["user"]["d"])  # design dimension
    p = np.int32(gen_specs["user"]["p"])  # objective dimension
    ub = np.asarray(gen_specs["user"]["ub"], dtype=ctypes.c_double, order="f")
    lb = np.asarray(gen_specs["user"]["lb"], dtype=ctypes.c_double, order="f")
    new_run = bool(gen_specs["user"]["new_run"])  # Are we starting a new run?
    isnb = np.int32(gen_specs["user"]["isnb"])  # first iter batch size
    snb = np.int32(gen_specs["user"]["snb"])  # search batch size
    onb = np.int32(gen_specs["user"]["onb"])  # preferred opt batch size
    # Check for any optional inputs
    lopt_budget = np.int32(2500)
    if "lopt_budget" in gen_specs["user"].keys():
        lopt_budget = gen_specs["user"]["lopt_budget"]
    decay = ctypes.c_double(0.5)
    if "decay" in gen_specs["user"].keys():
        decay = ctypes.c_double(gen_specs["user"]["decay"])
    eps = ctypes.c_double(np.sqrt(float_info.epsilon))
    if "eps" in gen_specs["user"].keys():
        eps = ctypes.c_double(gen_specs["user"]["eps"])
    des_tol = ctypes.c_double(np.sqrt(eps))
    if "des_tol" in gen_specs["user"].keys():
        des_tol = ctypes.c_double(gen_specs["user"]["des_tol"])
    epsw = ctypes.c_double(np.sqrt(np.sqrt(float_info.epsilon)))
    if "epsw" in gen_specs["user"].keys():
        epsw = ctypes.c_double(gen_specs["user"]["epsw"])
    obj_tol = ctypes.c_double(np.sqrt(eps))
    if "obj_tol" in gen_specs["user"].keys():
        obj_tol = ctypes.c_double(gen_specs["user"]["obj_tol"])
    obj_bounds = np.zeros((p, 2), dtype=ctypes.c_double, order="f")
    obj_bounds[:, 0] = -float_info.max
    obj_bounds[:, 1] = float_info.max
    if "obj_bounds" in gen_specs["user"].keys():
        obj_bounds = np.asarray(gen_specs["user"]["obj_bounds"], dtype=ctypes.c_double)
    trust_radf = ctypes.c_double(0.2)
    trust_radf_float = 0.2
    if "trust_radf" in gen_specs["user"].keys():
        trust_radf = ctypes.c_double(gen_specs["user"]["trust_radf"])
        trust_radf_float = float(gen_specs["user"]["trust_radf"])
    min_radf = ctypes.c_double(0.1 * trust_radf_float)
    if "min_radf" in gen_specs["user"].keys():
        min_radf = ctypes.c_double(gen_specs["user"]["min_radf"])
    pmode = bool(False)
    if "pmode" in gen_specs["user"].keys():
        pmode = bool(gen_specs["user"]["pmode"])

    # Copy the history arrays
    des_pts = np.asarray(H["x"].T, dtype=ctypes.c_double, order="f")
    obj_pts = np.asarray(H["f"].T, dtype=ctypes.c_double, order="f")

    # Is this the beginning of a new run? Reinitialize VTMOP
    if new_run:
        ierror = vtmop_libe_mod.vtmop_libe_init(d, p, lb, ub, lopt_budget, decay, des_tol, eps, epsw, obj_tol, min_radf,
                                                trust_radf, obj_bounds, pmode)
        if ierror != 0:
            return
        else:
            gen_specs["user"]["new_run"] = bool(False)

        # Get the first batch right away
        lbatch, batchx, ierr = vtmop_libe_mod.vtmop_libe_generate(d, p, lb, ub, des_pts, obj_pts, isnb, snb, onb,
                                                                  obj_bounds)
        # If lbatch is 0, immediately generate the second batch
        if lbatch == 0:
            lbatch, batchx, ierr = vtmop_libe_mod.vtmop_libe_generate(d, p, lb, ub, des_pts, obj_pts, isnb, snb, onb,
                                                                      obj_bounds)
    # Is this a resuming run? Just generate a batch of points
    else:
        # Generate a batch
        lbatch, batchx, ierr = vtmop_libe_mod.vtmop_libe_generate(d, p, lb, ub, des_pts, obj_pts, isnb, snb, onb,
                                                                  obj_bounds)

    # Read record
    Out = np.zeros(lbatch, dtype=gen_specs["out"])
    for i in range(0, lbatch):
        Out["x"][i, :] = batchx[:, i]

    return Out, persis_info



from libensemble.message_numbers import WORKER_DONE
from libensemble.executors.executor import Executor
import numpy as np


def circ_offset(wid, length):
    offset = length//2 + (length % 2 > 0)
    circ_index = wid + offset
    if circ_index > length:
        circ_index = circ_index - length
    return circ_index


def exp_nodelist_for_worker(exp_list, workerID, nodes_per_worker, persis_gens):
    """Modify expected node-lists based on workerID"""
    comps = exp_list.split()
    new_line = []
    for comp in comps:
        if comp.startswith('node-'):
            new_node_list = []
            node_list = comp.split(',')
            for node in node_list:
                node_name, node_num = node.split('-')
                offset = workerID - (1 + persis_gens)
                new_num = int(node_num) + int(nodes_per_worker*offset)
                new_node = '-'.join([node_name, str(new_num)])
                new_node_list.append(new_node)
            new_list = ','.join(new_node_list)
            new_line.append(new_list)
        else:
            new_line.append(comp)
    return ' '.join(new_line)


def runline_check(H, persis_info, sim_specs, libE_info):
    """Check run-lines produced by executor provided by a list of tests"""
    calc_status = 0
    x = H['x'][0][0]
    exctr = Executor.executor
    test_list = sim_specs['user']['tests']
    exp_list = sim_specs['user']['expect']
    npw = sim_specs['user']['nodes_per_worker']
    p_gens = sim_specs['user'].get('persis_gens', 0)

    for i, test in enumerate(test_list):
        task = exctr.submit(calc_type='sim',
                            num_procs=test.get('nprocs', None),
                            num_nodes=test.get('nnodes', None),
                            procs_per_node=test.get('ppn', None),
                            extra_args=test.get('e_args', None),
                            app_args='--testid ' + test.get('testid', None),
                            stdout='out.txt',
                            stderr='err.txt',
                            hyperthreads=test.get('ht', None),
                            dry_run=True)

        outline = task.runline
        new_exp_list = exp_nodelist_for_worker(exp_list[i], libE_info['workerID'], npw, p_gens)

        if outline != new_exp_list:
            print('outline is: {}\nexp     is: {}'.format(outline, new_exp_list), flush=True)

        assert(outline == new_exp_list)

    calc_status = WORKER_DONE
    output = np.zeros(1, dtype=sim_specs['out'])
    output['f'][0] = np.linalg.norm(x)
    return output, persis_info, calc_status


def runline_check_by_worker(H, persis_info, sim_specs, libE_info):
    """Check run-lines produced by executor provided by a list of lines per worker"""
    offset_for_min_rsets_scheduler = sim_specs['user'].get('offset_for_scheduler', False)
    calc_status = 0
    x = H['x'][0][0]
    exctr = Executor.executor
    test = sim_specs['user']['tests'][0]
    exp_list = sim_specs['user']['expect']
    p_gens = sim_specs['user'].get('persis_gens', 0)

    task = exctr.submit(calc_type='sim',
                        num_procs=test.get('nprocs', None),
                        num_nodes=test.get('nnodes', None),
                        procs_per_node=test.get('ppn', None),
                        extra_args=test.get('e_args', None),
                        app_args='--testid ' + test.get('testid', None),
                        stdout='out.txt',
                        stderr='err.txt',
                        hyperthreads=test.get('ht', None),
                        dry_run=True)

    outline = task.runline
    wid = libE_info['workerID']

    # Adjust for minimum slot scheduling in alloc func: e.g. 5 rsets on 2 nodes
    # Node 1: 3 rsets (0,1,2). Node 2: 2 rsets (3,4)
    # The first sim will got to rset 3 as it finds a "smaller slot".
    # Alternative would be for splitter to use opposite splits (e.g. 2,3 rather than 3,2)
    if offset_for_min_rsets_scheduler:
        wid_mod = circ_offset(wid, len(exp_list))
    else:
        wid_mod = wid

    new_exp_list = exp_list[wid_mod-1-p_gens]

    if outline != new_exp_list:
        print('Worker {}:\n outline is: {}\n exp     is: {}'.format(wid, outline, new_exp_list), flush=True)

    assert(outline == new_exp_list)

    calc_status = WORKER_DONE
    output = np.zeros(1, dtype=sim_specs['out'])
    output['f'][0] = np.linalg.norm(x)
    return output, persis_info, calc_status



"""
Example using NSGA2 as a libE generator function
For more about NSGA2, see
https://gist.github.com/darden1/fa8f96185a46796ed9516993bfe24862

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_deap_nsga2.py
   python3 test_deap_nsga2.py --nworkers 3 --comms local
   python3 test_deap_nsga2.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 3 4
# TESTSUITE_EXTRA: true

import numpy as np
from time import time
import os
from libensemble.libE import libE
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.sim_funcs.six_hump_camel import six_hump_camel_func
from libensemble.gen_funcs.persistent_deap_nsga2 import deap_nsga2 as gen_f

nworkers, is_manager, libE_specs, _ = parse_args()


def deap_six_hump(H, persis_info, sim_specs, _):
    xvals = H['individual'][0]
    Out = np.zeros(1, dtype=sim_specs['out'])
    y0 = np.linalg.norm(xvals)
    y1 = six_hump_camel_func(xvals)
    Out['fitness_values'] = (y0, y1)  # Requires tuple, even with 1 objective

    return Out, persis_info


if is_manager:
    start_time = time()

assert nworkers >= 2, "Cannot run with a persistent gen_f if only one worker."

# Number of generations, population size, individual size, and objectives
ngen = 125
pop_size = 100
ind_size = 2
num_obj = 2

# Variable Bounds (deap requires lists, not arrays!!!)
lb = [-3.0, -2.0]
ub = [3.0, 2.0]
w = (-1.0, -1.0)  # Must be a tuple

# State the objective function, its arguments, output, and necessary parameters (and their sizes)
sim_specs = {
    'sim_f': deap_six_hump,  # This is the function whose output is being minimized
    'in': ['individual'],  # These keys will be given to the above function
    'out': [('fitness_values', float, num_obj)],  # This output is being minimized
}  # end of sim spec

# State the generating function, its arguments, output, and necessary parameters.
gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id', 'generation', 'individual', 'fitness_values'],
    'persis_in': ['fitness_values', 'sim_id'],
    'out': [('individual', float, ind_size), ('generation', int), ('last_points', bool)],
    'user': {
        'lb': lb,
        'ub': ub,
        'weights': w,
        'pop_size': pop_size,
        'indiv_size': ind_size,
        'cxpb': 0.8,  # probability two individuals are crossed
        'eta': 20.0,  # large eta = low variation in children
        'indpb': 0.8 / ind_size,
    },  # end user
}  # end gen specs

alloc_specs = {'alloc_f': alloc_f, 'user': {'give_all_with_same_priority': True}}

# Tell libEnsemble when to stop
# 'sim_max' = number of simulation calls
# For deap, this should be pop_size*number of generations+1
exit_criteria = {'sim_max': pop_size * (ngen + 1)}

for run in range(3):

    persis_info = add_unique_random_streams({}, nworkers + 1)

    # Number of points in the sample
    num_samp = 100

    H0_dtype = [
        ('individual', float, ind_size),
        ('generation', int),
        ('fitness_values', float, num_obj),
        ('sim_id', int),
        ('returned', bool),
        ('given_back', bool),
        ('given', bool),
    ]
    H0 = np.zeros(num_samp, dtype=H0_dtype)

    # Mark these points as already have been given to be evaluated, and returned, but not given_back.
    H0['generation'][:] = 1
    # Give these points sim_ids
    H0['sim_id'] = range(num_samp)

    # "Load in" the points and their function values. (In this script, we are
    # actually evaluating them, but in many cases, they are available from past
    # evaluations
    np.random.seed(0)
    H0['individual'] = np.random.uniform(lb, ub, (num_samp, len(lb)))
    for i, x in enumerate(H0['individual']):
        H_dummy = np.zeros(1, dtype=[('individual', float, ind_size)])
        H_dummy['individual'] = x
        objs = deap_six_hump(H_dummy, {}, sim_specs, {})
        H0['fitness_values'][i] = objs[0]

    if run == 0:

        # Mark these points as already have been given to be evaluated, and returned, but not given_back.
        H0[['given', 'returned']] = True
        H0['given_back'][:] = False

    elif run == 1:
        H0[['given', 'returned', 'given_back']] = False

        # Testing use_persis_return_gen capabilities
        libE_specs['use_persis_return_gen'] = True

    else:
        H0 = None

    # Perform the run
    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs, H0=H0)

    if is_manager:
        if run == 0:
            assert np.sum(H['last_points']) == 0, (
                "The last_points shouldn't be marked (even though "
                "they were marked in the gen) as 'use_persis_return_gen' was false."
            )
        elif run == 1:
            assert np.sum(H['last_points']) == pop_size, (
                "The last_points should be marked as true because they were "
                "marked in the manager and 'use_persis_return_gen' is true."
            )

        script_name = os.path.splitext(os.path.basename(__file__))[0]
        assert flag == 0, script_name + " didn't exit correctly"
        assert sum(H['returned']) >= exit_criteria['sim_max'], script_name + " didn't evaluate the sim_max points."
        assert min(H['fitness_values'][:, 0]) <= 4e-3, script_name + " didn't find the minimum for objective 0."
        assert min(H['fitness_values'][:, 1]) <= -1.0, script_name + " didn't find the minimum for objective 1."
